/*****************************************************************************
 * Core A
 * TestProject.c
 *****************************************************************************/

#include <ccblkfn.h>
#include <stdio.h>
#include <string.h>

#include "emac.h"
#include "spif.h"
#include "tsc.h"

#define printf(...)

void pll_setup(unsigned vco_mul, unsigned cclk_div, unsigned sclk_div)
{
	unsigned int ca_iwr1 = *pSICA_IWR1;
	unsigned int ca_iwr0 = *pSICA_IWR0;
	
	
	*pSICA_IWR1 = 0x00;
	*pSICA_IWR0 = 0x01;
	
	/* Set PLL_CTL[MSEL] to vco_mul */
	unsigned int pll_ctl = *pPLL_CTL & ~(0x3f << 9);
	pll_ctl |= (vco_mul & 0x3f) << 9;
	*pPLL_CTL = pll_ctl;
	
	/* Set PLL_DIV[SSEL] to sclk_div, and find the appropriate cclk_div value */
	unsigned int cclk_val = 1;
	switch (cclk_div) {
	case 8:
		cclk_val = 0x3;
		break;
	case 4:
		cclk_val = 0x2;
		break;
	case 2:
		cclk_val = 0x1;
		break;
	case 1:
		cclk_val = 0x0;	
	}
	
	unsigned int pll_div = (sclk_div & 0xf) | (cclk_val << 4);
	
	*pPLL_DIV = pll_div;
	
	idle();

	*pSICA_IWR1 = ca_iwr1;
	*pSICA_IWR0 = ca_iwr0;
}

#define BUS_WIDTH_16_BITS		0x1
#define BUS_WIDTH_32_BITS		0x0

#define BANK_SIZE_64_MB			0x2
#define BANK_SIZE_128_MB		0x3

#define BANK_COL_ADDR_8			0x0
#define BANK_COL_ADDR_9			0x1
#define BANK_COL_ADDR_10		0x2
#define BANK_COL_ADDR_11		0x3

#define CDDBG_TRISTATE			(1 << 30)
#define PSSE_ENABLE				(1 << 23)

#define SDRAM_TWR(x) (((x) & 0x3) << 19)
#define SDRAM_TRCD(x) (((x) & 0x7) << 15)
#define SDRAM_TRP(x) (((x) & 0x7) << 11)
#define SDRAM_TRAS(x) (((x) & 0xf) <<6)
#define SDRAM_PASR(x) (((x) & 0x3) << 4)
#define SDRAM_CASL(x) (((x) & 0x3) << 2)

#define SDRAM_SCK1E_EN	(1 << 1)
#define SDRAM_SCTLE_EN  (1)

void async_setup_controller(void)
{
	*pEBIU_AMGCTL = 0x0002; /* Enable Bank 0, 32-bit bus */
	*pEBIU_AMBCTL0 = 0xFFC22416;
}

void sdram_setup_bank(int bank_id, int bank_size_mb, int bank_col_size)
{
	unsigned int b = ((bank_col_size & 0x3) << 4) |
					 ((bank_size_mb & 0x3) << 1)  |
					 1; /* Enable Bank */
	unsigned int r = *pEBIU_SDBCTL;

	r &= ~(0xff << ((bank_id) << 3));
	r |= (b & 0xff) << ((bank_id) << 3);
	
	*pEBIU_SDBCTL = r;
}

void sdram_setup_controller(int bus_width)
{
	unsigned int gc = 0;
	
	gc |= ((bus_width & 1) << 31) | CDDBG_TRISTATE | PSSE_ENABLE |
		SDRAM_TWR(0x2) | SDRAM_TRCD(2) | SDRAM_TRP(3) | SDRAM_TRAS(6) |
		SDRAM_CASL(3) | SCTLE;

	*pEBIU_SDRRC = 0x31A;		
	*pEBIU_SDGCTL = gc;

	/* Let's Go! */
	*(unsigned int *)0x1000;
}

unsigned int sdram_memtest(void)
{
	unsigned int i, j;
	for (i = 4096; i < (64 * 1024 * 1024); i+=4) {
		unsigned int *a = (unsigned int *)i;
		
		*a = i;	
	}
	
	for (j = 4096; j < (64 * 1024 * 1024); j+=4) {
		unsigned int *a = (unsigned int *)j;
		
		if (*a != j) {
			printf("[MT] [0x%08X] = 0x%08X\n", j, *a);
			return j;
		}
	}

	printf("[MT] 64MB RAM passed.\n");
		
	return 0;
}

#define PAGE_1KB		0x0
#define PAGE_4KB		0x1
#define PAGE_1MB		0x2
#define PAGE_4MB		0x3

#define DCPLB_VALID			0x1
#define DCPLB_SUPV_WR_EN	(0x1 << 4)

#define PAGE_SHIFT(x) ((x) << 16)

void set_cplb_entry(int entry_id, int page_size, unsigned int base_addr)
{
	volatile unsigned long *dcplb_data = pDCPLB_DATA0;
	volatile unsigned long *dcplb_addr = (volatile unsigned long *)pDCPLB_ADDR0;
	
	
	dcplb_addr[entry_id] = base_addr;
	dcplb_data[entry_id] = DCPLB_VALID | DCPLB_SUPV_WR_EN | PAGE_SHIFT(page_size);
}

#define SPI_CTL_SPE		(1 << 14)
#define SPI_CTL_MSTR	(1 << 12)

void setup_spi_flash(void)
{
	*pSPI_BAUD = 0x2; // 30 MHz
}

void read_flash_page(void)
{
	
}

int config_read = 0;
unsigned int c[256/sizeof(unsigned int)];

#define CONFIG_MAC_HI 0
#define CONFIG_MAC_LO 1

void get_configuration_word(unsigned int word)
{
	if (config_read == 0) {
		setup_spi_flash();
			
	}
}

static void flag_setup(void)
{
	*pFIO0_FLAG_D = 0x11ce;
	*pFIO0_DIR = 0xf7fe;
}

static unsigned short flash_spi_trans(unsigned short data, unsigned len)
{
	if(len == 2)
		*pSPI_CTL = 0x5109;
	else
		*pSPI_CTL = 0x5009;
	*pFIO0_FLAG_C = 0x0004;
	*pSPI_TDBR = data;
	while (!(*pSPI_STAT & 0x20));
	while (!(*pSPI_STAT & 0x1));
	return *pSPI_RDBR;
}

static inline void flash_spi_done(void)
{
	*pFIO0_FLAG_S = 0x0004;
}

static unsigned flash_spi_flash_query(void)
{
	unsigned id;

	/* WRDI */
	flash_spi_trans(0x04, 1);
	flash_spi_done();

	/* DBSY */
	flash_spi_trans(0x80, 1);
	flash_spi_done();

	flash_spi_trans(0x9000, 2);
	flash_spi_trans(0x0000, 2);
	id = flash_spi_trans(0x0000, 2);
	flash_spi_done();

	printf("DSP Flash ID = 0x%04X\n", id);
		
	return 0;
}

static void flash_spi_unprotect(void)
{
	/* EWSR */
	flash_spi_trans(0x50, 1);
	flash_spi_done();

	/* WRSR */
	flash_spi_trans(0x0100, 2);
	flash_spi_done();
}

static int flash_spi_64k_erase(unsigned base)
{
	/* WREN */
	flash_spi_trans(0x06, 1);
	flash_spi_done();

	/* 64KB BLOCK ERASE */
	flash_spi_trans(0xD800 | ((base >> 16) & 0xFF), 2);
	flash_spi_trans(base, 2);
	flash_spi_done();

	/* RDSR */
	flash_spi_trans(0x05, 1);
	while(flash_spi_trans(0x00, 1) & 1) ;
	flash_spi_done();

	return 0;	
}

static int flash_spi_256b_write(unsigned base, unsigned char *buf)
{
	unsigned i;

	/* WREN */
	flash_spi_trans(0x06, 1);
	flash_spi_done();

	/* 64KB BLOCK ERASE */
	flash_spi_trans(0x0200 | ((base >> 16) & 0xFF), 2);
	flash_spi_trans(base, 2);
	for(i=0; i<256; i+=2)
		flash_spi_trans((((unsigned short)buf[i]) << 8) | buf[i+1], 2);
	flash_spi_done();

	/* RDSR */
	flash_spi_trans(0x05, 1);
	while(flash_spi_trans(0x00, 1) & 1) ;
	flash_spi_done();

	return 0;	
}

static int flash_spi_verify(unsigned base, unsigned len, unsigned char *data)
{
	unsigned i, err = 0;
	unsigned char rd;

	/* FAST READ */
	flash_spi_trans(0x0B00 | ((base >> 16) & 0xFF), 2);
	flash_spi_trans(base, 2);
	flash_spi_trans(0x00, 1);
	for(i=0; i<len; i++) {
		rd = flash_spi_trans(0x00, 1);
		if(rd != data[i]) {
			printf("DSP SPI Flash: at %06X should be %02X, is %02X\n", base + i, data[i], rd);
			err ++;
			if(err > 10) {
				flash_spi_done();
				return err;
			}
		}
	}
	flash_spi_done();

	return err;	
}

static void flash_spi_readid(unsigned len, unsigned char *data)
{
	unsigned i;

	flash_spi_trans(0x8800, 2);
	for(i=0; i<len; i++)
		data[i] = flash_spi_trans(0x00, 1);
	flash_spi_done();
}

static unsigned fpga_spi_trans(unsigned data, unsigned len)
{
	*pSPIF_CTL = SPIF_CTL_ENABLE;
	*pSPIF_WRITE(len) = data;
	*pSPIF_CTL;
	while (*pSPIF_CTL);
	return *pSPIF_WRITE8;
}

static inline void fpga_spi_done(void)
{
	*pSPIF_CTL = SPIF_CTL_DISABLE;
}

static unsigned fpga_spi_flash_query(void)
{
	unsigned id;

	/* WRDI */
	fpga_spi_trans(0x04, 1);
	fpga_spi_done();

	/* DBSY */
	fpga_spi_trans(0x80, 1);
	fpga_spi_done();

	fpga_spi_trans(0x90000000, 4);
	id = fpga_spi_trans(0x0000, 2);
	fpga_spi_done();
	
	printf("FPGA Flash ID = 0x%04X\n", id);
		
	return 0;
}

static void fpga_spi_wait_bsy(void)
{
	*pSPIF_CTL = SPIF_CTL_ENABLE;
	*pSPIF_MISO;
	while(!*pSPIF_MISO);
	*pSPIF_CTL = SPIF_CTL_DISABLE;
}

static void fpga_spi_unprotect(void)
{
	/* EWSR */
	fpga_spi_trans(0x50, 1);
	fpga_spi_done();

	/* WRSR */
	fpga_spi_trans(0x0100, 2);
	fpga_spi_done();
}

static int fpga_spi_chip_erase(void)
{
	/* WREN */
	fpga_spi_trans(0x06, 1);
	fpga_spi_done();

	/* CHIP ERASE */
	fpga_spi_trans(0x60, 1);
	fpga_spi_done();

	/* RDSR */
	fpga_spi_trans(0x05, 1);
	while(fpga_spi_trans(0x00, 1) & 1) ;
	fpga_spi_done();

	return 0;	
}

static int fpga_spi_write(unsigned base, unsigned len, unsigned short *data)
{
	unsigned i;

	/* EBSY */
	fpga_spi_trans(0x70, 1);
	fpga_spi_done();

	/* WREN */
	fpga_spi_trans(0x06, 1);
	fpga_spi_done();

	/* AAI */
	fpga_spi_trans(0xAD000000 | base, 4);
	fpga_spi_trans((data[0] >> 8) | (data[0] << 8), 2);
	fpga_spi_done();

	fpga_spi_wait_bsy();

	for(i=1; i<len; i++) {
		fpga_spi_trans(0xAD0000 | (data[i] >> 8) | ((data[i] & 0xFF) << 8), 3);
		fpga_spi_done();

		fpga_spi_wait_bsy();
	}

	/* WRDI */
	fpga_spi_trans(0x04, 1);
	fpga_spi_done();

	/* DBSY */
	fpga_spi_trans(0x80, 1);
	fpga_spi_done();

	return 0;	
}

static int fpga_spi_verify(unsigned base, unsigned len, unsigned char *data)
{
	unsigned i, err = 0;
	unsigned char rd;

	/* FAST READ */
	fpga_spi_trans(0x0B000000 | base, 4);
	fpga_spi_trans(0x00, 1);
	for(i=0; i<len; i++) {
		rd = fpga_spi_trans(0x00, 1);
		if(rd != data[i]) {
			printf("FPGA SPI Flash: at %06X should be %02X, is %02X\n", base + i, data[i], rd);
			err ++;
			if(err > 10) {
				fpga_spi_done();
				return err;
			}
		}
	}
	fpga_spi_done();

	return err;	
}

/* experimental sensor handling */

#include "tscrom.h"

static void sensor_tsc_load(void)
{
	unsigned i;
	for(i=0; i<TSC_ROM_SIZE; i++)
		*pTSC_UC(i) = tsc_rom[i];
	*pTSC_CR = TSC_CR_RESET;
	*pTSC_CR = 0;
}

static void sensor_dac_spi_write(unsigned char command, unsigned short data)
{
	*pSPI_BAUD = 0x0020;
	*pSPI_CTL = 0x5809;
	*pFIO0_FLAG_C = 0x0040;
	*pSPI_TDBR = command;
	while (!(*pSPI_STAT & 0x20));
	while (!(*pSPI_STAT & 0x1));
	*pSPI_RDBR;
	*pSPI_CTL = 0x5909;
	*pSPI_TDBR = data;
	while (!(*pSPI_STAT & 0x20));
	while (!(*pSPI_STAT & 0x1));
	*pSPI_RDBR;
	*pFIO0_FLAG_S = 0x0040;
}

#define SENSOR_ADC0 0x10
#define SENSOR_ADC1 0x20
static void sensor_adc_spi_write(unsigned char sel, unsigned char reg, unsigned char data)
{
	*pSPI_BAUD = 0x0010;
	*pSPI_CTL = 0x5109;
	*pSPI_TDBR = (((unsigned short)reg) << 8) | data;
	while (!(*pSPI_STAT & 0x20));
	while (!(*pSPI_STAT & 0x1));
	*pSPI_RDBR;

	*pFIO0_FLAG_S = sel;
	*pTSC_SR; /* delay */
	*pTSC_SR;
	*pFIO0_FLAG_C = sel;
}

static void sensor_ibis_spi_write(unsigned short data)
{
	*pSPI_BAUD = 0x0010;
	*pSPI_CTL = 0x5109;
	*pFIO0_FLAG_C = 0x0008;
	*pSPI_TDBR = data;
	while (!(*pSPI_STAT & 0x20));
	while (!(*pSPI_STAT & 0x1));
	*pFIO0_FLAG_S = 0x0008;
	*pSPI_RDBR;
}

static unsigned short sensor_dac_setup[] = {
	0x28, 0x0001, /* full reset */
	0x20, 0x000F, /* power up all 4 DACs */
	0x38, 0x0001, /* reference on */
	0x00, 0x1999, /* write DAC A */
	0x01, 0x9999, /* write DAC B */
	0x02, 0xCCCC, /* write DAC C */
	0x13, 0xA8F5, /* write DAC D, update all */
	0xFF };

static unsigned char sensor_adc_setup[] = {
	SENSOR_ADC0 | SENSOR_ADC1, 0x04, 0x00,
	SENSOR_ADC0 | SENSOR_ADC1, 0x01, 0x28,
  	SENSOR_ADC0 | SENSOR_ADC1, 0x02, 0x14,
	SENSOR_ADC0 | SENSOR_ADC1, 0x03, 0x0D,
	SENSOR_ADC0 | SENSOR_ADC1, 0x06, 0x00,
	SENSOR_ADC0,               0x07, 0x04,
	              SENSOR_ADC1, 0x07, 0x14,
	SENSOR_ADC0 | SENSOR_ADC1, 0x08, 0x00,
	SENSOR_ADC0 | SENSOR_ADC1, 0x01, 0x29,
	SENSOR_ADC0 | SENSOR_ADC1, 0x27, 0x01,
	SENSOR_ADC0 | SENSOR_ADC1, 0x2B, 0x0B,
	0 };

static void sensor_poweron(void)
{
	unsigned i;

	/* load microcode to TSC */
	sensor_tsc_load();

	/* turn off sensor power and turn off ADC */
	*pFIO0_FLAG_C = 0x0400;
	sensor_adc_spi_write(SENSOR_ADC0 | SENSOR_ADC1, 0x01, 0x28);
	sensor_adc_spi_write(SENSOR_ADC0 | SENSOR_ADC1, 0x07, 0x14);

	/* turn on sensor power */
    for(i=0; sensor_dac_setup[i] != 0xFF; i+=2)
    	sensor_dac_spi_write(sensor_dac_setup[i], sensor_dac_setup[i+1]);
	*pFIO0_FLAG_S = 0x0400;
	*pTSC_CR = TSC_CR_POWER;
			
	for(i=0; sensor_adc_setup[i]; i+=3)
		sensor_adc_spi_write(sensor_adc_setup[i], sensor_adc_setup[i+1], sensor_adc_setup[i+2]);

	for(i=0; i<1048576; i++)
		*pTSC_CR;
		
	sensor_ibis_spi_write(0x7C45); /* sensor on, full resolution, RG/BG swizzle on */
}

static void sensor_frame_discard(void)
{
	*pPPI0_CONTROL = 0x018C;
	*pTSC_IP = 0x003;

	*pTSC_SR;
	*pTSC_SR;
	while(!(*pTSC_SR & TSC_SR_DONE));
}

static void sensor_frame_capture(void)
{
	unsigned *flags = (void *)0x00100000;
	unsigned i;

	/* setup 2D-DMA */
    *pDMA1_0_CURR_ADDR    = *pDMA1_0_START_ADDR = (void *)0x00100000;
                            *pDMA1_0_CONFIG     = 0x109A;
    *pDMA1_0_CURR_X_COUNT = *pDMA1_0_X_COUNT    = 0x02FB;
                            *pDMA1_0_X_MODIFY   = 0x0008;
    *pDMA1_0_CURR_Y_COUNT = *pDMA1_0_Y_COUNT    = 0x11D1;
                            *pDMA1_0_Y_MODIFY   = 0x0008;
                            *pDMA1_0_CONFIG     = 0x109B;

    *pDMA1_1_CURR_ADDR    = *pDMA1_1_START_ADDR = (void *)0x00100004;
                            *pDMA1_1_CONFIG     = 0x109A;
    *pDMA1_1_CURR_X_COUNT = *pDMA1_1_X_COUNT    = 0x02FB;
                            *pDMA1_1_X_MODIFY   = 0x0008;
    *pDMA1_1_CURR_Y_COUNT = *pDMA1_1_Y_COUNT    = 0x11D1;
                            *pDMA1_1_Y_MODIFY   = 0x0008;
                            *pDMA1_1_CONFIG     = 0x109B;

	/* setup PPI for 1-FS mode */
	*pPPI0_CONTROL = 0x018C;
	*pPPI0_DELAY   = 0x0000;
	*pPPI0_COUNT   = 0x0BEB;
	*pPPI0_FRAME   = 0x11D1;
//	*pPPI0_CONTROL = 0x018D;
	*pPPI0_CONTROL = 0x01FD;

	*pPPI1_CONTROL = 0x018C;
	*pPPI1_DELAY   = 0x0000;
	*pPPI1_COUNT   = 0x0BEB;
	*pPPI1_FRAME   = 0x11D1;
//	*pPPI1_CONTROL = 0x018D;
	*pPPI1_CONTROL = 0x01FD;

	/* start TSC microcode */
	*pTSC_IP = 0x003;

	*pTSC_SR;
	*pTSC_SR;
	while(!(*pTSC_SR & TSC_SR_DONE));
}

static void sensor_poweroff(void)
{
	*pTSC_CR = TSC_CR_RESET;
	*pTSC_CR = 0;
	sensor_adc_spi_write(SENSOR_ADC0 | SENSOR_ADC1, 0x01, 0x28);
	sensor_adc_spi_write(SENSOR_ADC0 | SENSOR_ADC1, 0x07, 0x14);
	*pFIO0_FLAG_C = 0x0400;
}

/* base Ethernet protocol */

#define LINK_EPROTO     0x6342

#define LINK_REQ_ENUM	0x0100
#define LINK_REQ_SYNC   0x0101
#define LINK_REQ_WBUF   0x2000
#define LINK_REQ_RBUF   0x2001
#define LINK_REQ_WBDF   0x2002
#define LINK_REQ_WBFF   0x2003
#define LINK_REQ_SONN   0x3000
#define LINK_REQ_SOFF   0x3001
#define LINK_REQ_SCAP   0x3002

#define LINK_REPLY      0x8000

#define LINK_SIZE_SHORT 24
#define LINK_SIZE_PAD   60
#define LINK_SIZE_LONG  280
struct link_pkt {
    unsigned char  dst[6];		/*  +0 */
    unsigned char  src[6];		/*  +6 */
    unsigned short eproto;		/* +12 */
    unsigned char  seq;			/* +14 */
    unsigned char  ack;			/* +15 */
    unsigned short req;			/* +16 */
    unsigned short _rsvd;		/* +18 */
    unsigned int   param;		/* +20 */
    unsigned char  data[256];	/* +24 */
} emac_pkt;

unsigned char emac_hwaddr[6] = { 0x00, 0x0C, 0xE9, 0xA5, 0xA5, 0xA5 }; /* Bloomberg LP :-) */
unsigned char emac_partner[6];
unsigned short emac_seq, emac_ack;

static const int link_speed[4] = { 1000, 1000, 100, 10 };
static void emac_early_setup(void)
{
	unsigned status = *pEMAC_RX_STAT;
	printf("EMAC variant ID: %07X, address: %02X:%02X:%02X:%02X:%02X:%02X\n", status >> 4,
		emac_hwaddr[0], emac_hwaddr[1], emac_hwaddr[2], emac_hwaddr[3], emac_hwaddr[4], emac_hwaddr[5]);
	printf("Link status: %s, %d Mbps, %s-duplex\n", (status & 1) ? "up" : "down", link_speed[(status & 6) >> 1], (status & 8) ? "full" : "half");
	
	*pEMAC_RX_MACL = ((unsigned *)emac_hwaddr)[0];
	*pEMAC_RX_MACL;
	*pEMAC_RX_MACH = ((unsigned short *)emac_hwaddr)[2] | EMAC_RXEN;
	*pEMAC_RX_MACH;
}

static void memcpy32(unsigned *dst, unsigned *src, unsigned num)
{
	unsigned i;
	for(i=0; i<num; i++) {
		*dst = *src;
		src ++;
		dst ++;
	}
}

static void emac_tx_pkt(int len)
{
	unsigned pktfree;

	while(*pEMAC_TX_FREE != 0xFFFFFFFF) ;

	*pEMAC_TX_START = 0x1000;
	*pEMAC_TX_LEN   = len | EMAC_TX_LAST;
	memcpy32((unsigned *)0x1000, (unsigned *)&emac_pkt, len >> 2);

	while((pktfree = *pEMAC_TX_FREE) == 0xFFFFFFFF) ;
}

static void emac_txrx_loop(void)
{
	int lastseq;
	
	memset(emac_pkt.dst, 0xFF,        6);
	memcpy(emac_pkt.src, emac_hwaddr, 6);
	emac_pkt.eproto = LINK_EPROTO;
	emac_pkt.req = 0;
	emac_pkt.seq = emac_pkt.ack = 0;
	emac_pkt.param = 0;
	emac_tx_pkt(LINK_SIZE_PAD);
	
	for (;;) {
		int bcast;
		unsigned len;
		while(!(len = *pEMAC_RX_POLL));
		if((*pEMAC_RX_DATA(12) & 0xFFFF) != LINK_EPROTO) {
			*pEMAC_RX_ACK = 0;
			*pEMAC_RX_POLL;
			continue;
		}

		if(len >= LINK_SIZE_LONG)
			len = LINK_SIZE_LONG;
		else
			len = LINK_SIZE_SHORT;
		memcpy(&emac_pkt, (void *)pEMAC_RX_DATA(0), len);
		*pEMAC_RX_ACK = 0;

		bcast = (emac_pkt.dst[0] == 0xFF) && (emac_pkt.dst[1] == 0xFF) &&
				(emac_pkt.dst[2] == 0xFF) && (emac_pkt.dst[3] == 0xFF) && 
				(emac_pkt.dst[4] == 0xFF) && (emac_pkt.dst[5] == 0xFF);
		
		memcpy(emac_pkt.dst, emac_pkt.src, 6);
		memcpy(emac_pkt.src, emac_hwaddr,  6);
		emac_pkt.req |= LINK_REPLY;

		if(emac_pkt.req == (LINK_REQ_ENUM|LINK_REPLY)) {
			emac_pkt.ack = 0;
			emac_pkt.seq = 0;
			emac_tx_pkt(LINK_SIZE_PAD);
			continue;
		}

/*
		if(emac_pkt.seq == lastseq)
			printf("Tx packet dropped (%d)!\n", lastseq);
*/
		lastseq = emac_pkt.seq;

		if(bcast)
			continue;

		if(emac_pkt.req == (LINK_REQ_SYNC|LINK_REPLY)) {
			memcpy(emac_partner, emac_pkt.dst, 6);
			emac_pkt.ack = emac_pkt.seq;
			emac_pkt.seq ^= *(unsigned short *)emac_pkt.dst;
			emac_tx_pkt(LINK_SIZE_PAD);
			printf("Connected to %02X:%02X:%02X:%02X:%02X:%02X\n", emac_pkt.dst[0], emac_pkt.dst[1],
				emac_pkt.dst[2], emac_pkt.dst[3], emac_pkt.dst[4], emac_pkt.dst[5]);
			continue;
		}

		if(memcmp(emac_pkt.dst, emac_partner, 6))
			continue;

		emac_pkt.ack = emac_pkt.seq;
		emac_pkt.seq = emac_pkt.ack + 1;

		if(emac_pkt.req == (LINK_REQ_WBUF|LINK_REPLY) && len == LINK_SIZE_LONG) {
			memcpy((void *)(0x00100000 + emac_pkt.param), emac_pkt.data, 256);
			emac_tx_pkt(LINK_SIZE_PAD);
			continue;
		}
		if(emac_pkt.req == (LINK_REQ_RBUF|LINK_REPLY)) {
			memcpy(emac_pkt.data, (void *)(0x00100000 + emac_pkt.param), 256);
			emac_tx_pkt(LINK_SIZE_LONG);
			continue;
		}
	
		if(emac_pkt.req == (LINK_REQ_WBFF|LINK_REPLY)) {
			emac_tx_pkt(LINK_SIZE_PAD);

			*pFIO0_FLAG_C = 0x7000;
			*pFIO0_FLAG_S = 0x8000;
			printf("FPGA SPI Flash: Chip Erase\n");
			fpga_spi_unprotect();
			fpga_spi_chip_erase();
			printf("FPGA SPI Flash: Program\n");
			fpga_spi_write(0, emac_pkt.param >> 1, (void *)0x00100000);
			printf("FPGA SPI Flash: Verify\n");
			if(fpga_spi_verify(0, emac_pkt.param, (void *)0x00100000)) {
				printf("FPGA SPI Flash: Programming failed\n");
				*pFIO0_FLAG_T = 0xC000;
			} else {
				printf("FPGA SPI Flash: Success\n");
				*pFIO0_FLAG_T = 0x9000;
			}

			continue;
		}

		if(emac_pkt.req == (LINK_REQ_WBDF|LINK_REPLY)) {
			unsigned addr;

			emac_tx_pkt(LINK_SIZE_PAD);

			*pFIO0_FLAG_C = 0x7000;
			*pFIO0_FLAG_S = 0x8000;
			flash_spi_unprotect();
			printf("DSP SPI Flash: Erase & program\n");
			for(addr=0; addr<emac_pkt.param; addr += 256) {
				if(!(addr & 0xFFFF))
					flash_spi_64k_erase(addr);
				flash_spi_256b_write(addr, (void *)(0x00100000 + addr));
			}
			printf("DSP SPI Flash: Verify\n");
			if(flash_spi_verify(0, emac_pkt.param, (void *)0x00100000)) {
				printf("DSP SPI Flash: Programming failed\n");
				*pFIO0_FLAG_T = 0xC000;
			} else {
				printf("DSP SPI Flash: Success\n");
				*pFIO0_FLAG_T = 0x9000;
			}

			continue;
		}

		if(emac_pkt.req == (LINK_REQ_SONN|LINK_REPLY)) {
			*pFIO0_FLAG_S = 0x2000;
			sensor_poweron();
			emac_tx_pkt(LINK_SIZE_PAD);
			continue;
		}

		if(emac_pkt.req == (LINK_REQ_SOFF|LINK_REPLY)) {
			*pFIO0_FLAG_C = 0x2000;
			sensor_poweroff();
			emac_tx_pkt(LINK_SIZE_PAD);
			continue;
		}

		if(emac_pkt.req == (LINK_REQ_SCAP|LINK_REPLY)) {
			*pFIO0_FLAG_S = 0x4000;
			sensor_frame_discard();
			*pFIO0_FLAG_S = 0x8000;
			sensor_frame_capture();
			*pFIO0_FLAG_C = 0xC000;
			emac_tx_pkt(LINK_SIZE_PAD);
			continue;
		}
				
		printf("Connection packet <%04X>\n", emac_pkt.req & ~LINK_REPLY);
	}
}

int main( void )
{
	adi_core_b_enable();

	flag_setup();
	
	/* Take 25MHz -> 600 MHz, set CoreClock divider to 1, and SysClk to 5 */
	pll_setup(24, 1, 5);

	/* Setup Bank 0 */
	async_setup_controller();	
		
	/* Setup Bank 0 */
	sdram_setup_bank(0, BANK_SIZE_64_MB, BANK_COL_ADDR_9);
	sdram_setup_controller(BUS_WIDTH_32_BITS);

	sensor_poweroff();
		
	set_cplb_entry(0, PAGE_4MB, 0x00000000); // Add translation for base of SDRAM
	set_cplb_entry(1, PAGE_4MB, 0x00400000);
	set_cplb_entry(2, PAGE_4MB, 0x00800000);
	set_cplb_entry(3, PAGE_4MB, 0x00C00000);
	set_cplb_entry(4, PAGE_4MB, 0x01000000);
	set_cplb_entry(5, PAGE_4MB, 0x01400000);
	set_cplb_entry(6, PAGE_4MB, 0x01800000);
	set_cplb_entry(7, PAGE_4MB, 0x01C00000);
	set_cplb_entry(8, PAGE_4MB, 0x20000000); // Add translation for EMAC
	
	unsigned int b = 0;
	if ( (b = sdram_memtest()) != 0) {
		printf("Memory failed at address 0x%08x.\n", b);
	}
	
	printf("Hello, world at 600 MHz\n");
	
	setup_spi_flash();
	flash_spi_flash_query();
	flash_spi_readid(3, emac_hwaddr + 3);

	fpga_spi_flash_query();

	emac_early_setup();
	emac_txrx_loop();

	return 0;
}
