#include <ccblkfn.h>
#include <stdio.h>
#include <string.h>

#include <cpu_spi.h>

void local_flag_setup(void)
{
	*pFIO0_FLAG_D = 0x11ce;
	*pFIO0_DIR = 0xf7fe;
	*pFIO0_INEN = 0x0800;
}

static unsigned short flash_spi_trans(unsigned short data, unsigned len)
{
	*pSPI_BAUD = 0x2; /* 30 MHz */
	if(len == 2)
		*pSPI_CTL = 0x5109;
	else
		*pSPI_CTL = 0x5009;
	*pFIO0_FLAG_C = 0x0004;
	*pSPI_TDBR = data;
	while (!(*pSPI_STAT & 0x20));
	while (!(*pSPI_STAT & 0x1));
	return *pSPI_RDBR;
}

static inline void flash_spi_done(void)
{
	*pFIO0_FLAG_S = 0x0004;
}

unsigned flash_spi_flash_query(void)
{
	unsigned id;

	/* WRDI */
	flash_spi_trans(0x04, 1);
	flash_spi_done();

	/* DBSY */
	flash_spi_trans(0x80, 1);
	flash_spi_done();

	flash_spi_trans(0x9000, 2);
	flash_spi_trans(0x0000, 2);
	id = flash_spi_trans(0x0000, 2);
	flash_spi_done();

	printf("DSP Flash ID = 0x%04X\n", id);
		
	return 0;
}

void flash_spi_unprotect(void)
{
	/* EWSR */
	flash_spi_trans(0x50, 1);
	flash_spi_done();

	/* WRSR */
	flash_spi_trans(0x0100, 2);
	flash_spi_done();
}

int flash_spi_64k_erase(unsigned base)
{
	/* WREN */
	flash_spi_trans(0x06, 1);
	flash_spi_done();

	/* 64KB BLOCK ERASE */
	flash_spi_trans(0xD800 | ((base >> 16) & 0xFF), 2);
	flash_spi_trans(base, 2);
	flash_spi_done();

	/* RDSR */
	flash_spi_trans(0x05, 1);
	while(flash_spi_trans(0x00, 1) & 1) ;
	flash_spi_done();

	return 0;	
}

int flash_spi_256b_write(unsigned base, unsigned char *buf)
{
	unsigned i;

	/* WREN */
	flash_spi_trans(0x06, 1);
	flash_spi_done();

	/* 64KB BLOCK ERASE */
	flash_spi_trans(0x0200 | ((base >> 16) & 0xFF), 2);
	flash_spi_trans(base, 2);
	for(i=0; i<256; i+=2)
		flash_spi_trans((((unsigned short)buf[i]) << 8) | buf[i+1], 2);
	flash_spi_done();

	/* RDSR */
	flash_spi_trans(0x05, 1);
	while(flash_spi_trans(0x00, 1) & 1) ;
	flash_spi_done();

	return 0;	
}

int flash_spi_verify(unsigned base, unsigned len, unsigned char *data)
{
	unsigned i, err = 0;
	unsigned char rd;

	/* FAST READ */
	flash_spi_trans(0x0B00 | ((base >> 16) & 0xFF), 2);
	flash_spi_trans(base, 2);
	flash_spi_trans(0x00, 1);
	for(i=0; i<len; i++) {
		rd = flash_spi_trans(0x00, 1);
		if(rd != data[i]) {
			printf("DSP SPI Flash: at %06X should be %02X, is %02X\n", base + i, data[i], rd);
			err ++;
			if(err > 10) {
				flash_spi_done();
				return err;
			}
		}
	}
	flash_spi_done();

	return err;	
}

void flash_spi_readid(unsigned len, unsigned char *data)
{
	unsigned i;

	flash_spi_trans(0x8800, 2);
	for(i=0; i<len; i++)
		data[i] = flash_spi_trans(0x00, 1);
	flash_spi_done();
}
