#include <stdio.h>
#include <string.h>
#include <sys/exception.h>
#include <services/services.h>
#include <fpga_irq.h>

#define FPGA_IRQ         0x20044000
#define pFPGA_IRQ        ((volatile unsigned int *)FPGA_IRQ)

static int initialized = 0;
static struct {
	firq_handler handler;
	void *ptr;
} irq_handlers[8];

unsigned char firq_get_status(void)
{
	return (*pFPGA_IRQ) & (*pFPGA_IRQ >> 16);
}

void firq_enable(u8 irq)
{
	(*pFPGA_IRQ) |= (0x10000 << irq);
}

void firq_disable(u8 irq)
{
	(*pFPGA_IRQ) &= ~(0x10000 << irq);
}

ADI_INT_HANDLER(firq_handler_fn)
{
	int i;
	u8 stat;

	stat = firq_get_status();
	
	if(!stat)
		return ADI_INT_RESULT_NOT_PROCESSED;

	for(i=0; i<8; i++)
		if(stat & (1 << i)) {
			if(irq_handlers[i].handler)
				irq_handlers[i].handler(i, irq_handlers[i].ptr);
			else {
				printf("Unhandled IRQ %d\n", i);
				firq_disable(i);
				/* unhandled fIRQ */
			}
		}
		
	return ADI_INT_RESULT_PROCESSED;
}

void firq_set_handler(unsigned char irq, firq_handler handler, void *ptr)
{
	int ivg = ADI_INT_IVG_9;
	
	if(!initialized) {
		memset(irq_handlers, 0, sizeof(irq_handlers));

 	    adi_int_CECHook(ivg, firq_handler_fn, NULL, FALSE);
    	adi_int_SICSetIVG(ADI_INT_PF0_15_A, ivg);
    	adi_int_SICEnable(ADI_INT_PF0_15_A);

    	/* enable FIRQ line from FPGA (PF11) */
    	*pFIO0_MASKA_S = (1 << 11);
    			
		initialized = 1;
	}
	
	irq_handlers[irq].handler = handler;
	irq_handlers[irq].ptr = ptr;
}
