module bigcam_etx(clk125,
    m_brn, m_bgn, m_bghn, m_firq,
    m_clk0, m_asn0, m_oen, m_awen, m_ren,
    m_cke, m_csn0, m_rasn, m_casn, m_wen, m_sa10, m_ba, m_a, m_dm, m_d,
    e_txclk, e_txen, e_txd, e_rxclk, e_rxdv, e_rxd,
    bf_cen, bf_sck, bf_mosi, bf_miso,
    t_select, t_clk_yl, t_reset, t_syl, t_sync_yl, t_sync_yr, t_clk_y, t_syr, t_sync_y, t_clk_yr, t_sync_x, t_clk_x, t_pc, t_shr, t_shs,
    t_mclk, t_rsmp, t_vsmp,
    t_ppiclk, t_ppihs, t_ppivs);

input         clk125;

output        m_brn;
input         m_bgn;
input         m_bghn;
output        m_firq;

inout         m_clk0;
input         m_asn0;
input         m_oen;
input         m_awen;
input         m_ren;

inout         m_cke;
inout         m_csn0;
inout         m_rasn;
inout         m_casn;
inout         m_wen;
inout         m_sa10;
inout   [1:0] m_ba;
inout  [12:0] m_a;
inout   [3:0] m_dm;
inout  [31:0] m_d;

output        e_txclk;
output        e_txen;
output  [3:0] e_txd;

input         e_rxclk;
input         e_rxdv;
input   [3:0] e_rxd;

output        bf_cen;
output        bf_sck;
output        bf_mosi;
input         bf_miso;

output        t_select;
output        t_clk_yl;
output        t_reset;
output        t_syl;
output        t_sync_yl;
output        t_sync_yr;
output        t_clk_y;
output        t_syr;
output        t_sync_y;
output        t_clk_yr;
output        t_sync_x;
output        t_clk_x;
output        t_pc;
output        t_shr;
output        t_shs;

output        t_mclk;
output        t_rsmp;
output        t_vsmp;

output        t_ppiclk;
output        t_ppihs;
output        t_ppivs;

wire          i_clk0;
wire          clk125_b;
wire          eclk;
IBUFG clk125_ibufg (.I(clk125),  .O(clk125_b));
BUFGMUX  eclk_bufg (.I0(i_clk0), .I1(clk125_b), .S(1'b1), .O(eclk));

reg           mrst = 1'b1;
reg     [9:0] mrst_cntr = 10'h3FF;
always @(posedge eclk) begin
    if(mrst_cntr)
        mrst_cntr <= mrst_cntr - 10'h001;
    mrst <= |mrst_cntr;
end

wire          clk0;

wire          wr_vld;
wire    [1:0] wr_type;
wire   [23:0] wr_addr;
wire   [31:0] wr_data;

wire          sdreq_vld;
wire   [11:0] sdreq_count;
wire   [23:0] sdreq_addr;

wire          rd_dvld_tx, rd_dvld_rx, rd_dvld_bf, rd_dvld_st;
wire   [31:0] rd_data_tx, rd_data_rx, rd_data_bf, rd_data_st;

wire          rd_vld;
wire   [14:0] rd_addr;
wire          rd_dvld = rd_dvld_tx || rd_dvld_rx || rd_dvld_bf || rd_dvld_st;
wire   [31:0] rd_data = rd_dvld_st ? rd_data_st : rd_dvld_bf ? rd_data_bf : rd_dvld_tx ? rd_data_tx : rd_data_rx;

wire          irq_txmgr_free;
wire          irq_rxmac_ready;
wire          irq_link_change;
wire          irq_stim_done;

mcif umcif (eclk, mrst, clk0, i_clk0, clk125_b,
    m_brn, m_bgn, m_bghn, m_firq,
    m_clk0, m_asn0, m_oen, m_awen, m_ren,
    m_cke, m_csn0, m_rasn, m_casn, m_wen, m_sa10, m_ba, m_a, m_dm, m_d,
    wr_vld, wr_type, wr_addr, wr_data,
    sdreq_vld, sdreq_count, sdreq_addr,
    rd_vld, rd_addr, rd_dvld, rd_data,
    { 4'h0, irq_stim_done, irq_link_change, irq_rxmac_ready, irq_txmgr_free });

wire          tx_dvld;
wire   [31:0] tx_data;

wire          tx_pvld;
wire   [13:0] tx_plen;
wire    [1:0] tx_pflg;

wire          cr_vld;
wire   [11:0] cr_count;

txmgr utxmgr (clk0, mrst,
    wr_vld, wr_type, wr_addr, wr_data,
    sdreq_vld, sdreq_count, sdreq_addr,
    rd_vld, rd_addr, rd_dvld_tx, rd_data_tx,
    tx_dvld, tx_data,
    tx_pvld, tx_plen, tx_pflg,
    cr_vld, cr_count,
    irq_txmgr_free);

wire          link_up;
wire    [1:0] ck_div;
wire          full_duplex;

txmac utxmac (clk0, mrst, eclk, mrst,
    ck_div,
    tx_dvld, tx_data,
    tx_pvld, tx_plen, tx_pflg,
    cr_vld, cr_count,
    e_txclk, e_txen, e_txd);

rxmac urxmac (clk0, mrst,
    link_up, ck_div, full_duplex,
    wr_vld, wr_type, wr_addr, wr_data,
    rd_vld, rd_addr, rd_dvld_rx, rd_data_rx,
    e_rxclk, e_rxdv, e_rxd,
    irq_rxmac_ready, irq_link_change);

spif uspif (clk0, mrst,
    wr_vld, wr_type, wr_addr, wr_data,
    rd_vld, rd_addr, rd_dvld_bf, rd_data_bf,
    bf_cen, bf_sck, bf_mosi, bf_miso);

stim ustim (clk0, mrst, eclk, mrst,
    wr_vld, wr_type, wr_addr, wr_data,
    rd_vld, rd_addr, rd_dvld_st, rd_data_st,
    t_select, t_clk_yl, t_reset, t_syl, t_sync_yl, t_sync_yr, t_clk_y, t_syr, t_sync_y, t_clk_yr, t_sync_x, t_clk_x, t_pc, t_shr, t_shs,
    t_mclk, t_rsmp, t_vsmp,
    t_ppiclk, t_ppihs, t_ppivs,
    irq_stim_done);

endmodule
