`define WR_TYPE_ASYNC	2'b00
`define WR_TYPE_SDRAM   2'b01
`define WR_TYPE_RDATA   2'b11

module mcif(mclk, mrst, clk0, i_clk0, i_mclk,
    m_brn, m_bgn, m_bghn, m_firq,
    m_clk0, m_asn0, m_oen, m_awen, m_ren,
    m_cke, m_csn0, m_rasn, m_casn, m_wen, m_sa10, m_ba, m_a, m_dm, m_d,
    wr_vld, wr_type, wr_addr, wr_data,
    sdreq_vld, sdreq_count, sdreq_addr,
    rd_vld, rd_addr, rd_dvld, rd_data,
    irq_vld);

input         mclk;
input         mrst;
output        clk0;
output        i_clk0;
input         i_mclk;

output        m_brn;
input         m_bgn;
input         m_bghn;
output        m_firq;

inout         m_clk0;
input         m_asn0;
input         m_oen;
input         m_awen;
input         m_ren;

inout         m_cke;
inout         m_csn0;
inout         m_rasn;
inout         m_casn;
inout         m_wen;
inout         m_sa10;
inout   [1:0] m_ba;
inout  [12:0] m_a;
inout   [3:0] m_dm;
inout  [31:0] m_d;

output        wr_vld;
output  [1:0] wr_type;
output [23:0] wr_addr;
output [31:0] wr_data;

input         sdreq_vld;
input  [11:0] sdreq_count;
input  [23:0] sdreq_addr;

output        rd_vld;
output [14:0] rd_addr;
input         rd_dvld;
input  [31:0] rd_data;

input   [7:0] irq_vld;

/* I/O pads */

reg          o_brn_n;
wire         o_brn = o_brn_n || mrst;
wire         n_bgn;
wire         i_bghn;
wire         o_firq;

OBUF brn_obuf (.O(m_brn), .I(o_brn));
IBUF bgn_ibuf (.I(m_bgn), .O(n_bgn)); 
PULLUP bgn_pull (.O(m_bgn));
IBUF bghn_ibuf (.I(m_bghn), .O(i_bghn)); 
OBUF firq_obuf (.O(m_firq), .I(o_firq));

wire         o_clk0;
wire         clk0;
wire         n_asn0;
wire         n_oen;
wire         n_awen;
wire         n_ren;

reg          t_clk;
reg          t_bus;
wire         t_clk0;

ODDR2 clk0_oddr  (.Q(o_clk0), .C0(mclk), .C1(!mclk), .CE(1'b1), .D0(!t_clk), .D1(1'b0),  .R(1'b0), .S(1'b0));
ODDR2 clk0_oddrt (.Q(t_clk0), .C0(mclk), .C1(!mclk), .CE(1'b1), .D0(t_clk),  .D1(t_clk), .R(1'b0), .S(1'b0));
IOBUF clk0_iobuf (.IO(m_clk0), .I(o_clk0), .O(i_clk0), .T(t_clk0));
BUFGMUX clk0_bufg (.I0(i_mclk), .I1(i_clk0), .S(t_clk || mrst), .O(clk0));
IBUF asn0_ibuf (.I(m_asn0), .O(n_asn0));
PULLUP asn0_pull (.O(m_asn0));
IBUF oen_ibuf (.I(m_oen), .O(n_oen));
IBUF awen_ibuf (.I(m_awen), .O(n_awen));
IBUF ren_ibuf (.I(m_ren), .O(n_ren));

wire         n_cke;
wire         n_csn0;
wire         n_rasn;
wire         n_casn;
wire         n_wen;
wire         n_sa10;
wire   [1:0] n_ba;
wire  [12:0] n_a;
wire   [3:0] n_dm;
wire  [31:0] n_d;

reg          o_cke;
reg          o_csn0;
reg          o_rasn;
reg          o_casn;
reg          o_wen;
reg          o_sa10;
reg    [1:0] o_ba;
reg   [12:0] o_a;
reg    [3:0] o_dm;
reg   [31:0] o_d;

wire         t_d = n_oen || n_asn0;

IOBUF cke_iobuf (.IO(m_cke), .I(o_cke), .O(n_cke), .T(t_bus));
IOBUF csn0_iobuf (.IO(m_csn0), .I(o_csn0), .O(n_csn0), .T(t_bus));
IOBUF rasn_iobuf (.IO(m_rasn), .I(o_rasn), .O(n_rasn), .T(t_bus));
IOBUF casn_iobuf (.IO(m_casn), .I(o_casn), .O(n_casn), .T(t_bus));
IOBUF wen_iobuf (.IO(m_wen), .I(o_wen), .O(n_wen), .T(t_bus));
IOBUF sa10_iobuf (.IO(m_sa10), .I(o_sa10), .O(n_sa10), .T(t_bus));
IOBUF ba0_iobuf (.IO(m_ba[0]), .I(o_ba[0]), .O(n_ba[0]), .T(t_bus));
IOBUF ba1_iobuf (.IO(m_ba[1]), .I(o_ba[1]), .O(n_ba[1]), .T(t_bus));
IOBUF a0_iobuf (.IO(m_a[0]), .I(o_a[0]), .O(n_a[0]), .T(t_bus));
IOBUF a1_iobuf (.IO(m_a[1]), .I(o_a[1]), .O(n_a[1]), .T(t_bus));
IOBUF a2_iobuf (.IO(m_a[2]), .I(o_a[2]), .O(n_a[2]), .T(t_bus));
IOBUF a3_iobuf (.IO(m_a[3]), .I(o_a[3]), .O(n_a[3]), .T(t_bus));
IOBUF a4_iobuf (.IO(m_a[4]), .I(o_a[4]), .O(n_a[4]), .T(t_bus));
IOBUF a5_iobuf (.IO(m_a[5]), .I(o_a[5]), .O(n_a[5]), .T(t_bus));
IOBUF a6_iobuf (.IO(m_a[6]), .I(o_a[6]), .O(n_a[6]), .T(t_bus));
IOBUF a7_iobuf (.IO(m_a[7]), .I(o_a[7]), .O(n_a[7]), .T(t_bus));
IOBUF a8_iobuf (.IO(m_a[8]), .I(o_a[8]), .O(n_a[8]), .T(t_bus));
IOBUF a9_iobuf (.IO(m_a[9]), .I(o_a[9]), .O(n_a[9]), .T(t_bus));
IOBUF a10_iobuf (.IO(m_a[10]), .I(o_a[10]), .O(n_a[10]), .T(t_bus));
IOBUF a11_iobuf (.IO(m_a[11]), .I(o_a[11]), .O(n_a[11]), .T(t_bus));
IOBUF a12_iobuf (.IO(m_a[12]), .I(o_a[12]), .O(n_a[12]), .T(t_bus));
IOBUF dm0_iobuf (.IO(m_dm[0]), .I(o_dm[0]), .O(n_dm[0]), .T(t_bus));
IOBUF dm1_iobuf (.IO(m_dm[1]), .I(o_dm[1]), .O(n_dm[1]), .T(t_bus));
IOBUF dm2_iobuf (.IO(m_dm[2]), .I(o_dm[2]), .O(n_dm[2]), .T(t_bus));
IOBUF dm3_iobuf (.IO(m_dm[3]), .I(o_dm[3]), .O(n_dm[3]), .T(t_bus));
IOBUF d0_iobuf (.IO(m_d[0]), .I(o_d[0]), .O(n_d[0]), .T(t_d));
IOBUF d1_iobuf (.IO(m_d[1]), .I(o_d[1]), .O(n_d[1]), .T(t_d));
IOBUF d2_iobuf (.IO(m_d[2]), .I(o_d[2]), .O(n_d[2]), .T(t_d));
IOBUF d3_iobuf (.IO(m_d[3]), .I(o_d[3]), .O(n_d[3]), .T(t_d));
IOBUF d4_iobuf (.IO(m_d[4]), .I(o_d[4]), .O(n_d[4]), .T(t_d));
IOBUF d5_iobuf (.IO(m_d[5]), .I(o_d[5]), .O(n_d[5]), .T(t_d));
IOBUF d6_iobuf (.IO(m_d[6]), .I(o_d[6]), .O(n_d[6]), .T(t_d));
IOBUF d7_iobuf (.IO(m_d[7]), .I(o_d[7]), .O(n_d[7]), .T(t_d));
IOBUF d8_iobuf (.IO(m_d[8]), .I(o_d[8]), .O(n_d[8]), .T(t_d));
IOBUF d9_iobuf (.IO(m_d[9]), .I(o_d[9]), .O(n_d[9]), .T(t_d));
IOBUF d10_iobuf (.IO(m_d[10]), .I(o_d[10]), .O(n_d[10]), .T(t_d));
IOBUF d11_iobuf (.IO(m_d[11]), .I(o_d[11]), .O(n_d[11]), .T(t_d));
IOBUF d12_iobuf (.IO(m_d[12]), .I(o_d[12]), .O(n_d[12]), .T(t_d));
IOBUF d13_iobuf (.IO(m_d[13]), .I(o_d[13]), .O(n_d[13]), .T(t_d));
IOBUF d14_iobuf (.IO(m_d[14]), .I(o_d[14]), .O(n_d[14]), .T(t_d));
IOBUF d15_iobuf (.IO(m_d[15]), .I(o_d[15]), .O(n_d[15]), .T(t_d));
IOBUF d16_iobuf (.IO(m_d[16]), .I(o_d[16]), .O(n_d[16]), .T(t_d));
IOBUF d17_iobuf (.IO(m_d[17]), .I(o_d[17]), .O(n_d[17]), .T(t_d));
IOBUF d18_iobuf (.IO(m_d[18]), .I(o_d[18]), .O(n_d[18]), .T(t_d));
IOBUF d19_iobuf (.IO(m_d[19]), .I(o_d[19]), .O(n_d[19]), .T(t_d));
IOBUF d20_iobuf (.IO(m_d[20]), .I(o_d[20]), .O(n_d[20]), .T(t_d));
IOBUF d21_iobuf (.IO(m_d[21]), .I(o_d[21]), .O(n_d[21]), .T(t_d));
IOBUF d22_iobuf (.IO(m_d[22]), .I(o_d[22]), .O(n_d[22]), .T(t_d));
IOBUF d23_iobuf (.IO(m_d[23]), .I(o_d[23]), .O(n_d[23]), .T(t_d));
IOBUF d24_iobuf (.IO(m_d[24]), .I(o_d[24]), .O(n_d[24]), .T(t_d));
IOBUF d25_iobuf (.IO(m_d[25]), .I(o_d[25]), .O(n_d[25]), .T(t_d));
IOBUF d26_iobuf (.IO(m_d[26]), .I(o_d[26]), .O(n_d[26]), .T(t_d));
IOBUF d27_iobuf (.IO(m_d[27]), .I(o_d[27]), .O(n_d[27]), .T(t_d));
IOBUF d28_iobuf (.IO(m_d[28]), .I(o_d[28]), .O(n_d[28]), .T(t_d));
IOBUF d29_iobuf (.IO(m_d[29]), .I(o_d[29]), .O(n_d[29]), .T(t_d));
IOBUF d30_iobuf (.IO(m_d[30]), .I(o_d[30]), .O(n_d[30]), .T(t_d));
IOBUF d31_iobuf (.IO(m_d[31]), .I(o_d[31]), .O(n_d[31]), .T(t_d));

/* register all inputs */

reg          i_bgn;
reg          i_asn0;
reg          i_oen;
reg          i_awen;
reg          i_ren;
reg          i_cke;
reg          i_csn0;
reg          i_rasn;
reg          i_casn;
reg          i_wen;
reg          i_sa10;
reg    [1:0] i_ba;
reg   [12:0] i_a;
reg    [3:0] i_dm;
reg   [31:0] i_d;

always @(posedge mclk)
    i_bgn  <= n_bgn;

always @(posedge clk0) begin
    i_asn0 <= n_asn0;
    i_oen  <= n_oen;
    i_awen <= n_awen;
    i_ren  <= n_ren;
    i_cke  <= n_cke;
    i_csn0 <= n_csn0;
    i_rasn <= n_rasn;
    i_casn <= n_casn;
    i_wen  <= n_wen;
    i_sa10 <= n_sa10;
    i_ba   <= n_ba;
    i_a    <= n_a;
    i_dm   <= n_dm;
    i_d    <= n_d;
end

/* SDRAM state tracker */

reg   [3:0] bank_open;
reg  [12:0] bank0_row;
reg  [12:0] bank1_row;
reg  [12:0] bank2_row;
reg  [12:0] bank3_row;

// tRRD = 2 clk
// tRCD = 3 clk
// tRP  = 3 clk
// tRAS = 7 clk
reg   [1:0] trrd;
reg   [1:0] trcd;
reg   [1:0] trp;
reg   [2:0] tras;

wire        trrd_ok = !trrd;
wire        trcd_ok = !trcd;
wire        trp_ok  = !trp;
wire        tras_ok = !tras;

always @(posedge clk0) begin
    if(!trrd_ok)
	trrd <= trrd - 2'b01;
    if(!trcd_ok)
	trcd <= trcd - 2'b01;
    if(!trp_ok)
	trp  <= trp  - 2'b01;
    if(!tras_ok)
	tras <= tras - 3'b001;

    if(i_cke && !i_csn0) begin
	/* ACTIVATE */
	if(!i_rasn && i_casn && i_wen) begin
	    trcd <= 2'b11;
	    trrd <= 2'b10;

	    if(i_ba == 2'b00) begin
		bank_open[0] <= 1'b1;
		bank0_row <= { i_a[12:11], i_sa10, i_a[9:0] };
	    end
	    if(i_ba == 2'b01) begin
		bank_open[1] <= 1'b1;
		bank1_row <= { i_a[12:11], i_sa10, i_a[9:0] };
	    end
	    if(i_ba == 2'b10) begin
		bank_open[2] <= 1'b1;
		bank2_row <= { i_a[12:11], i_sa10, i_a[9:0] };
	    end
	    if(i_ba == 2'b11) begin
		bank_open[3] <= 1'b1;
		bank3_row <= { i_a[12:11], i_sa10, i_a[9:0] };
	    end
	end

	/* PRECHARGE */
	if(!i_rasn && i_casn && !i_wen) begin
	    trp <= 2'b11;
	    tras <= 3'b111;

	    if(i_ba == 2'b00 || i_sa10)
		bank_open[0] <= 1'b0;
	    if(i_ba == 2'b01 || i_sa10)
		bank_open[1] <= 1'b0;
	    if(i_ba == 2'b10 || i_sa10)
		bank_open[2] <= 1'b0;
	    if(i_ba == 2'b11 || i_sa10)
		bank_open[3] <= 1'b0;
	end

	/* READ/WRITE WITH AP */
	if(i_rasn && !i_casn && i_sa10) begin
	    if(i_ba == 2'b00)
		bank_open[0] <= 1'b0;
	    if(i_ba == 2'b01)
		bank_open[1] <= 1'b0;
	    if(i_ba == 2'b10)
		bank_open[2] <= 1'b0;
	    if(i_ba == 2'b11)
		bank_open[3] <= 1'b0;
	end
    end

    if(mrst) begin
	bank_open <= 4'h0;
	trrd <= 2'b00;
	trcd <= 2'b00;
	trp  <= 2'b00;
	tras <= 3'b000;
    end
end

reg         sdr_rd, sdr_rd_q, sdr_rd_qq, sdr_rd_q3, sdr_rd_q4;
reg         sdr_last, sdr_last_q, sdr_last_qq, sdr_last_q3, sdr_last_q4;

/* data capture */

reg         wr_vld;
reg   [1:0] wr_type;
reg  [23:0] wr_addr;
reg  [31:0] wr_data;

reg         aw_state;

always @(posedge clk0) begin
    wr_vld <= 1'b0;

    if(i_cke && !i_csn0 && i_rasn && !i_casn && !i_wen && i_bgn) begin
	wr_vld <= 1'b1;
	wr_type <= `WR_TYPE_SDRAM;
	wr_addr[8:0] <= i_a[8:0];
	wr_addr[21:9] <= (i_ba == 2'b00) ? bank0_row : (i_ba == 2'b01) ? bank1_row : (i_ba == 2'b10) ? bank2_row : bank3_row;
	wr_addr[23:22] <= i_ba;
	wr_data <= i_d;
    end

    aw_state <= !i_asn0 && !i_awen;
    if(!i_asn0 && !i_awen && !aw_state) begin
	wr_vld <= 1'b1;
	wr_type <= `WR_TYPE_ASYNC;
	wr_addr[12:0] <= i_a;
	wr_addr[15:13] <= 3'b000;
	wr_addr[17:16] <= i_ba;
	wr_addr[23:18] <= 6'h00;
	wr_data <= i_d;
    end

    if(sdr_rd_q3) begin
	wr_vld  <= 1'b1;
	wr_type <= `WR_TYPE_RDATA;
	wr_addr <= { sdr_last_q3, 23'h000000 };
	wr_data <= i_d;
    end
end

reg         rd_vld;
reg  [14:0] rd_addr;

wire        rd_irq_vld;
wire [31:0] rd_irq_data;

reg         ar_state;

always @(posedge clk0) begin
    rd_vld <= 1'b0;

    ar_state <= !i_asn0 && !i_oen;
    if(!i_asn0 && !i_oen && !ar_state) begin
	rd_vld         <= 1'b1;
	rd_addr[14:13] <= i_ba;
	rd_addr[12:0]  <= i_a;
    end

    if(rd_dvld || rd_irq_vld)
	o_d <= rd_irq_vld ? rd_irq_data : rd_data;
end

/* SDRAM read state machine */

`define SDR_IDLE 2'h0
`define SDR_BREQ 2'h1
`define SDR_OPEN 2'h2
`define SDR_READ 2'h3

reg   [1:0] sdr_state;
reg  [11:0] sdr_word_cntr;
reg  [23:0] sdr_word_addr;

reg         sdr_idle_q, sdr_idle_qq;

wire        sdb_vld = (sdr_state == `SDR_OPEN);
wire  [1:0] sdb_bank = sdr_word_addr[23:22];
wire [12:0] sdb_row = sdr_word_addr[21:9];

wire        sdb_sel_open = (sdb_bank == 2'b00) ? bank_open[0] : (sdb_bank == 2'b01) ? bank_open[1] : (sdb_bank == 2'b10) ? bank_open[2] : bank_open[3];
wire [12:0] sdb_sel_row = (sdb_bank == 2'b00) ? bank0_row : (sdb_bank == 2'b01) ? bank1_row : (sdb_bank == 2'b10) ? bank2_row : bank3_row;

wire        sdb_sel_row_eq = (sdb_sel_row == sdb_row);
wire        sdb_sel_row_ok = sdb_sel_open && sdb_sel_row_eq;

reg         sdrdd_vld;
reg  [31:0] sdrdd_data;

reg         o_rasn_q;

always @(posedge mclk)
    t_clk <= n_bgn;

always @(posedge clk0) begin
    sdr_idle_q <= (sdr_state == `SDR_IDLE);
    sdr_idle_qq <= sdr_idle_q;
    o_brn_n <= (sdr_state == `SDR_IDLE) && sdr_idle_q && sdr_idle_qq;

    if(sdreq_vld) begin
	sdr_state <= `SDR_BREQ;
	sdr_word_cntr <= sdreq_count;
	sdr_word_addr <= sdreq_addr;
    end

    if(sdr_state == `SDR_BREQ && !i_bgn) begin
	sdr_state <= `SDR_OPEN;
	t_bus     <= 1'b0;
    end

    sdr_last <= 1'b0;
    if(sdr_state == `SDR_READ) begin
	sdr_word_addr <= sdr_word_addr + 24'h000001;
	sdr_word_cntr <= sdr_word_cntr - 12'h001;
	if(sdr_word_addr[8:0] == 9'h1FF)
	    sdr_state <= `SDR_OPEN;
	if(sdr_word_cntr == 12'h001) begin
	    sdr_state <= `SDR_IDLE;
	    sdr_last  <= 1'b1;
	end
    end

    if(sdr_last_qq)
        t_bus <= 1'b1;

    o_cke  <= 1'b1;
    o_csn0 <= 1'b1;
    o_rasn <= 1'b1;
    o_casn <= 1'b1;
    o_wen  <= 1'b1;

    o_rasn_q <= o_rasn;

    sdr_rd <= 1'b0;
    if(sdr_state == `SDR_READ) begin
	/* READ NO AP */
	o_csn0   <= 1'b0;
	o_casn   <= 1'b0;
	o_ba     <= sdb_bank;
	o_sa10   <= 1'b0;
	o_a[8:0] <= sdr_word_addr[8:0];
	sdr_rd   <= 1'b1;
    end

    if(sdb_vld && sdb_sel_open && !sdb_sel_row_eq && tras_ok && o_rasn && o_rasn_q) begin
	/* PRECHARGE */
	o_csn0 <= 1'b0;
	o_rasn <= 1'b0;
	o_wen  <= 1'b0;
	o_ba   <= sdb_bank;
	o_sa10 <= 1'b0;
    end

    if(sdb_vld && !sdb_sel_open && trrd_ok && trp_ok && o_rasn && o_rasn_q) begin
	/* ACTIVATE */
	o_csn0     <= 1'b0;
	o_rasn     <= 1'b0;
	o_ba       <= sdb_bank;
	o_a[9:0]   <= sdb_row[9:0];
	o_sa10     <= sdb_row[10];
	o_a[12:11] <= sdb_row[12:11];
    end

    sdr_rd_q    <= sdr_rd;
    sdr_rd_qq   <= sdr_rd_q;
    sdr_rd_q3   <= sdr_rd_qq;
    sdr_rd_q4   <= sdr_rd_q3;

    sdr_last_q  <= sdr_last;
    sdr_last_qq <= sdr_last_q;
    sdr_last_q3 <= sdr_last_qq;
    sdr_last_q4 <= sdr_last_q3;

    if(sdb_vld && sdb_sel_row_ok && trcd_ok && o_rasn && o_rasn_q)
	sdr_state <= `SDR_READ;

    if(mrst) begin
	sdr_state <= `SDR_IDLE;
	t_bus <= 1'b1;
    end
end

/* IRQ controller */

reg   [7:0] irq_mask, irq_vld_q;

assign      o_firq = |(irq_mask & irq_vld_q);

wire        wr_en_irq = wr_vld && (wr_type == `WR_TYPE_ASYNC) && (wr_addr[17:16] == 2'b01) && wr_addr[12];

assign      rd_irq_vld = rd_vld && (rd_addr[14:12] == 3'b011);
assign      rd_irq_data = { 8'h0, irq_mask, 8'h0, irq_vld_q };

always @(posedge clk0) begin
    if(wr_en_irq)
	irq_mask <= wr_data[23:16];
    irq_vld_q <= irq_vld;
end

endmodule
