`define WR_TYPE_ASYNC   2'b00
`define WR_TYPE_SDRAM   2'b01
`define WR_TYPE_RDATA   2'b11

`define CKDIV_1000 2'b00
`define CKDIV_100  2'b10
`define CKDIV_10   2'b11

module rxmac(clk0, mrst,
    link_up, ck_div, full_duplex,
    wr_vld, wr_type, wr_addr, wr_data,
    rd_vld, rd_addr, rd_dvld, rd_data,
    e_rxclk, e_rxdv, e_rxd,
    irq_ready, irq_link);

input         clk0;
input         mrst;

output  [1:0] ck_div;
output        link_up;
output        full_duplex;

input         wr_vld;
input   [1:0] wr_type;
input  [23:0] wr_addr;
input  [31:0] wr_data;

input         rd_vld;
input  [14:0] rd_addr;
output        rd_dvld;
output [31:0] rd_data;

input         e_rxclk;
input         e_rxdv;
input   [3:0] e_rxd;

output        irq_ready;
output        irq_link;

wire          e_rxclk_b;
wire          rclk;
IBUFG rxclk_ibuf (.I(e_rxclk),   .O(e_rxclk_b));
BUFG   rclk_bufg (.I(e_rxclk_b), .O(rclk));

reg     [3:0] rrstcntr = 4'hF;
reg           rrst = 1'b1;
always @(posedge rclk) begin
    if(rrstcntr)
        rrstcntr <= rrstcntr - 4'h1;
    else
        rrst <= 1'b0;
end

wire          e_rxdv_b;
wire    [3:0] e_rxd_b;
IBUF   rxdv_ibuf (.I(e_rxdv),    .O(e_rxdv_b));
IBUF   rxd0_ibuf (.I(e_rxd[0]),  .O(e_rxd_b[0]));
IBUF   rxd1_ibuf (.I(e_rxd[1]),  .O(e_rxd_b[1]));
IBUF   rxd2_ibuf (.I(e_rxd[2]),  .O(e_rxd_b[2]));
IBUF   rxd3_ibuf (.I(e_rxd[3]),  .O(e_rxd_b[3]));

wire          d_rxdv;
wire          d_rxdvl;
wire          d_rxctl;
wire          d_rxer = d_rxdv ^ d_rxctl;
wire    [7:0] d_rxd;
wire    [3:0] d_rxdl;
IDDR2 rxdv_iddr2 (.Q0(d_rxdvl),   .Q1(d_rxctl),  .C0(rclk), .C1(!rclk), .CE(1'b1), .D(e_rxdv_b),   .R(1'b0), .S(1'b0));
IDDR2 rxd0_iddr2 (.Q0(d_rxdl[0]), .Q1(d_rxd[4]), .C0(rclk), .C1(!rclk), .CE(1'b1), .D(e_rxd_b[0]), .R(1'b0), .S(1'b0));
IDDR2 rxd1_iddr2 (.Q0(d_rxdl[1]), .Q1(d_rxd[5]), .C0(rclk), .C1(!rclk), .CE(1'b1), .D(e_rxd_b[1]), .R(1'b0), .S(1'b0));
IDDR2 rxd2_iddr2 (.Q0(d_rxdl[2]), .Q1(d_rxd[6]), .C0(rclk), .C1(!rclk), .CE(1'b1), .D(e_rxd_b[2]), .R(1'b0), .S(1'b0));
IDDR2 rxd3_iddr2 (.Q0(d_rxdl[3]), .Q1(d_rxd[7]), .C0(rclk), .C1(!rclk), .CE(1'b1), .D(e_rxd_b[3]), .R(1'b0), .S(1'b0));
defparam rxdv_iddr2.DDR_ALIGNMENT = "C0";
defparam rxd0_iddr2.DDR_ALIGNMENT = "C0";
defparam rxd1_iddr2.DDR_ALIGNMENT = "C0";
defparam rxd2_iddr2.DDR_ALIGNMENT = "C0";
defparam rxd3_iddr2.DDR_ALIGNMENT = "C0";
defparam rxdv_iddr2.SRTYPE = "ASYNC";
defparam rxd0_iddr2.SRTYPE = "ASYNC";
defparam rxd1_iddr2.SRTYPE = "ASYNC";
defparam rxd2_iddr2.SRTYPE = "ASYNC";
defparam rxd3_iddr2.SRTYPE = "ASYNC";
FD rxdv_fd (.C(rclk), .D(d_rxdvl),   .Q(d_rxdv));
FD rxd0_fd (.C(rclk), .D(d_rxdl[0]), .Q(d_rxd[0]));
FD rxd1_fd (.C(rclk), .D(d_rxdl[1]), .Q(d_rxd[1]));
FD rxd2_fd (.C(rclk), .D(d_rxdl[2]), .Q(d_rxd[2]));
FD rxd3_fd (.C(rclk), .D(d_rxdl[3]), .Q(d_rxd[3]));

reg           link_up;
reg     [1:0] ck_div;
reg           full_duplex;

reg           m_rxen = 1'b0;
reg           m_rxpre = 1'b0;
reg           m_rxdv;
reg           m_rxer;
reg     [7:0] m_rxd;
always @(posedge rclk) begin
    if(!d_rxdv && !d_rxer && (d_rxd[2:1] != 2'b11)) begin
        link_up <= d_rxd[0];
        if(d_rxd[0]) begin
            ck_div      <= (d_rxd[2:1] == 2'b00) ? `CKDIV_10 : (d_rxd[2:1] == 2'b01) ? `CKDIV_100 : `CKDIV_1000;
            full_duplex <= d_rxd[3];
        end
    end

    m_rxen <= (d_rxdv && !m_rxdv) ? 1'b0 : ~m_rxen;
    m_rxdv <= d_rxdv;
    m_rxer <= d_rxer;

    m_rxd[3:0] <= m_rxd[7:4];
    m_rxd[7:4] <= d_rxd[3:0];

    if(d_rxdv && m_rxpre && (d_rxd[3:0] == 4'hD)) begin
        m_rxen  <= 1'b1;
        m_rxpre <= 1'b0;
    end
    if(d_rxdv && !m_rxdv)
        m_rxpre <= 1'b1;
    if(!d_rxdv)
        m_rxpre <= 1'b0;

    if(!ck_div[1]) begin
        m_rxen <= 1'b1;
        m_rxd  <= d_rxd;
    end

    if(rrst)
        ck_div <= `CKDIV_1000;
end

reg           promisc;
reg           rcven;
reg           arwe;
reg     [2:0] arwaddr;
reg    [31:0] arwd;
always @(posedge clk0) begin
    if(wr_vld && (wr_type == `WR_TYPE_ASYNC) && (wr_addr[17:16] == 2'b00) && (wr_addr[12:11] == 2'b01) && (wr_addr[2:0] == 2'b010)) begin
        arwe    <= 1'b1;
        arwaddr <= 3'b000;
        arwd    <= wr_data;
    end
    if(wr_vld && (wr_type == `WR_TYPE_ASYNC) && (wr_addr[17:16] == 2'b00) && (wr_addr[12:11] == 2'b01) && (wr_addr[2:0] == 2'b011)) begin
        promisc <= wr_data[30];
        rcven   <= wr_data[31];
        arwe    <= 1'b1;
        arwaddr <= 3'b100;
        arwd    <= wr_data;
    end

    if(arwe && (arwaddr == 3'b011))
        arwe    <= 1'b0;
    if(arwe && (arwaddr == 3'b101))
        arwe    <= 1'b0;
    if(arwe) begin
        arwaddr <= arwaddr + 3'b001;
        arwd    <= { 8'h00, arwd[31:8] };
    end

    if(mrst) begin
        promisc <= 1'b0;
        rcven   <= 1'b0;
        arwe    <= 1'b0;
    end
end

reg           start, data;
reg     [2:0] araddr;

wire    [7:0] ardata;
xram8x8 addram (.clk(clk0), .we(arwe), .wa(arwaddr), .wd(arwd[7:0]), .ra(araddr), .rd(ardata));
reg           ar_match;
reg           ar_bcast;
reg           ar_latch;
reg           noerr;

always @(posedge rclk)
    if(m_rxen) begin
        if(m_rxd != 8'hFF)
            ar_bcast <= 1'b0;
        if(m_rxd != ardata)
            ar_match <= 1'b0;
        if(m_rxdv && m_rxer)
            noerr    <= 1'b0;

        if(m_rxdv && !start && !data)
            start    <= 1'b1;
        if(m_rxdv && start && (m_rxd == 8'hD5)) begin
            start    <= 1'b0;
            data     <= 1'b1;
            araddr   <= 3'b000;
            ar_match <= 1'b1;
            ar_bcast <= 1'b1;
            ar_latch <= 1'b0;
            noerr    <= 1'b1;
        end

        if(araddr == 3'b110 && m_rxdv)
            ar_latch <= (ar_bcast || ar_match || promisc) && rcven;

        if(araddr != 3'b111)
            araddr <= araddr + 3'b001;

        if(!m_rxdv) begin
            start  <= 1'b0;
            data   <= 1'b0;
            araddr <= 3'b111;
        end
    end

wire   [31:0] crc;
crc32_8 rxcrc (.clk(rclk), .rst(start), .en(data && m_rxen), .i(m_rxd), .q(crc));
wire    [7:0] fcs0 = ~{ crc[24], crc[25], crc[26], crc[27], crc[28], crc[29], crc[30], crc[31] };
wire    [7:0] fcs1 = ~{ crc[16], crc[17], crc[18], crc[19], crc[20], crc[21], crc[22], crc[23] };
wire    [7:0] fcs2 = ~{ crc[8],  crc[9],  crc[10], crc[11], crc[12], crc[13], crc[14], crc[15] };
wire    [7:0] fcs3 = ~{ crc[0],  crc[1],  crc[2],  crc[3],  crc[4],  crc[5],  crc[6],  crc[7]  };
reg     [7:0] fcs1_q;
reg     [7:0] fcs2_q, fcs2_qq;
reg     [7:0] fcs3_q, fcs3_qq, fcs3_q3;
wire          fcs0eq = (fcs0    == m_rxd);
wire          fcs1eq = (fcs1_q  == m_rxd);
wire          fcs2eq = (fcs2_qq == m_rxd);
wire          fcs3eq = (fcs3_q3 == m_rxd);
reg           fcs0eq_q, fcs0eq_qq, fcs0eq_q3;
reg           fcs1eq_q, fcs1eq_qq;
reg           fcs2eq_q;
reg           fcsok;
always @(posedge rclk)
    if(m_rxen) begin
        fcs1_q    <= fcs1;
        fcs2_q    <= fcs2;
        fcs3_q    <= fcs3;
        fcs2_qq   <= fcs2_q;
        fcs3_qq   <= fcs3_q;
        fcs3_q3   <= fcs3_qq;
        fcs0eq_q  <= fcs0eq;
        fcs0eq_qq <= fcs0eq_q;
        fcs0eq_q3 <= fcs0eq_qq;
        fcs1eq_q  <= fcs1eq;
        fcs1eq_qq <= fcs1eq_q;
        fcs2eq_q  <= fcs2eq;
        fcsok     <= (fcs0eq_q3 && fcs1eq_qq && fcs2eq_q && fcs3eq);
    end

wire          pktack = data && !m_rxdv &&  (ar_latch && fcsok && noerr);
wire          pktnak = data && !m_rxdv && !(ar_latch && fcsok && noerr);

reg    [12:0] pkptr;
reg    [12:0] dwptr, dwptr_q;
reg    [12:0] pktlen;
always @(posedge rclk) begin
    if(m_rxen) begin
        if(start) begin
            pkptr <= dwptr;
            pktlen <= 13'h1FFF;
        end

        dwptr_q <= dwptr;
        if(data && m_rxdv) begin
            dwptr <= dwptr + 13'h0001;
            pktlen <= pktlen + 13'h0001;
        end

        if(pktack)
            dwptr <= { dwptr_q[12:2], 2'b00 };
        if(pktnak)
            dwptr <= pkptr;
    end

    if(rrst)
        dwptr <= 13'h0000;
end

reg    [10:0] drptr;

wire          dram_re = rd_vld && !rd_addr[14:13] && rd_addr[12] && !rd_addr[11];
wire   [10:0] dram_ra = rd_addr[10:0] + drptr;
wire   [31:0] dram_rd;
xram8192x8r32qa dram (.wclk(rclk), .we(data && m_rxdv && m_rxen), .wa(dwptr), .wd(m_rxd), .rclk(clk0), .re(dram_re), .ra(dram_ra), .rd(dram_rd));

wire          pfifo_rv;
wire   [15:0] pfifo_rd;
reg           pfifo_re;
xfifo512x16a pfifo (.wclk(rclk), .wrst(rrst), .we(pktack && m_rxen), .wd({ 3'b000, pktlen }), .rclk(clk0), .rrst(mrst), .re(pfifo_re), .rv(pfifo_rv), .rd(pfifo_rd));

wire    [3:0] link_stat = { full_duplex, ck_div, link_up };
reg     [3:0] link_ack;
reg           irq_link;

reg           rd_dvld;
reg           rd_dsel;
reg    [31:0] rd_data_n;
always @(posedge clk0) begin
    rd_dvld   <= rd_vld && !rd_addr[14:13] && (rd_addr[12] ^ rd_addr[11]);
    if(rd_vld) begin
        rd_dsel   <= rd_addr[12];
        rd_data_n <= rd_addr[2:1] ? { 28'hBCEA010, link_stat } : (pfifo_rv ? { 19'h00000, pfifo_rd[12:0] - 13'h0003 } : 32'h00000000);
    end

    if(wr_vld && (wr_type == `WR_TYPE_ASYNC) && (wr_addr[17:16] == 2'b00) && (wr_addr[12:11] == 2'b01) && (wr_addr[2:0] == 2'b100))
	link_ack <= wr_data[3:0];

    pfifo_re  <= pfifo_rv && wr_vld && (wr_type == `WR_TYPE_ASYNC) && (wr_addr[17:16] == 2'b00) && (wr_addr[12:11] == 2'b01) && (wr_addr[2:0] == 2'b001);
    if(pfifo_re)
        drptr <= drptr + pfifo_rd[12:2];

    irq_link <= (link_ack != link_stat);

    if(mrst) begin
        drptr <= 11'h000;
	link_ack <= 4'b0000;
    end
end

wire   [31:0] rd_data = rd_dsel ? dram_rd : rd_data_n;

wire          irq_ready = pfifo_rv;

endmodule
