/*
   6000 - write to send 8 LSBs out, read to get last input data
   6004 - write to send 16 LSBs out, read to get last input data
   6008 - write to send 24 LSBs out, read to get last input data
   600C - write to send all 32 bits out, read to get last input data
   6010 - write 0 to deselect flash, write 1 to select; read to get status (0 - idle, nonzero - busy)
 */

module spif (clk, rst,
    wr_vld, wr_type, wr_addr, wr_data,
    rd_vld, rd_addr, rd_dvld, rd_data,
    bf_cen, bf_sck, bf_mosi, bf_miso);

input         clk;
input         rst;

input         wr_vld;
input   [1:0] wr_type;
input  [23:0] wr_addr;
input  [31:0] wr_data;

input         rd_vld;
input  [14:0] rd_addr;
output        rd_dvld;
output [31:0] rd_data;

output        bf_cen;
output        bf_sck;
output        bf_mosi;
input         bf_miso;

reg           bf_cen;
reg           bf_sck;
reg           bf_mosi;

reg     [7:0] clkdiv;
reg    [31:0] data;

reg           rd_dvld;
reg    [31:0] rd_data;

always @(posedge clk) begin
    if(wr_vld && (wr_type == 2'b00) && (wr_addr[17:16] == 2'b00) && (wr_addr[12:11] == 2'b11) && !wr_addr[2]) begin
        clkdiv <= { wr_addr[1:0], 6'h3F };
        data[7:0]   <= (wr_addr[1:0] == 2'b11) ? wr_data[7:0]   : 8'h00;
        data[15:8]  <= (wr_addr[1:0] == 2'b11) ? wr_data[15:8]  : (wr_addr[1:0] == 2'b10) ? wr_data[7:0]   : 8'h00;
        data[23:16] <= (wr_addr[1:0] == 2'b11) ? wr_data[23:16] : (wr_addr[1:0] == 2'b10) ? wr_data[15:8]  : (wr_addr[1:0] == 2'b01) ? wr_data[7:0]  : 8'h00;
        data[31:24] <= (wr_addr[1:0] == 2'b11) ? wr_data[31:24] : (wr_addr[1:0] == 2'b10) ? wr_data[23:16] : (wr_addr[1:0] == 2'b01) ? wr_data[15:8] : wr_data[7:0];
    end

    if(clkdiv)
        clkdiv <= clkdiv - 8'h01;
    if(clkdiv[2:0] == 3'b010)
        data <= { data[30:0], bf_miso };

    bf_sck  <= clkdiv[1] ^ clkdiv[2];
    bf_mosi <= data[31];

    if(wr_vld && (wr_type == 2'b00) && (wr_addr[17:16] == 2'b00) && (wr_addr[12:11] == 2'b11) && wr_addr[2])
        bf_cen <= !wr_data[0];

    rd_dvld <= 1'b0;
    if(rd_vld && (rd_addr[14:11] == 4'b0011)) begin
        rd_dvld <= 1'b1;
        rd_data <= rd_addr[2] ? { 24'h000000, rd_addr[0] ? { 7'h00, bf_miso } : clkdiv } : data;
    end

    if(rst) begin
        bf_cen  <= 1'b1;
        bf_sck  <= 1'b0;
        bf_mosi <= 1'b0;
        clkdiv  <= 8'h00;
    end
end

endmodule

