`define WR_TYPE_ASYNC   2'b00
`define WR_TYPE_SDRAM   2'b01
`define WR_TYPE_RDATA   2'b11

module stim (clk0, mrst, eclk, erst,
    wr_vld, wr_type, wr_addr, wr_data,
    rd_vld, rd_addr, rd_dvld, rd_data,
    t_select, t_clk_yl, t_reset, t_syl, t_sync_yl, t_sync_yr, t_clk_y, t_syr, t_sync_y, t_clk_yr, t_sync_x, t_clk_x, t_pc, t_shr, t_shs,
    t_mclk, t_rsmp, t_vsmp,
    t_ppiclk, t_ppihs, t_ppivs,
    irq_done);

input         clk0;
input         mrst;
input         eclk;
input         erst;

input         wr_vld;
input   [1:0] wr_type;
input  [23:0] wr_addr;
input  [31:0] wr_data;

input         rd_vld;
input  [14:0] rd_addr;
output        rd_dvld;
output [31:0] rd_data;

output        t_select;
output        t_clk_yl;
output        t_reset;
output        t_syl;
output        t_sync_yl;
output        t_sync_yr;
output        t_clk_y;
output        t_syr;
output        t_sync_y;
output        t_clk_yr;
output        t_sync_x;
output        t_clk_x;
output        t_pc;
output        t_shr;
output        t_shs;

output        t_mclk;
output        t_rsmp;
output        t_vsmp;

output        t_ppiclk;
output        t_ppihs;
output        t_ppivs;

output        irq_done;

wire          wr_en = wr_vld && (wr_type == `WR_TYPE_ASYNC) && (wr_addr[17:16] == 2'b01);
wire          wr_en_uc = wr_en && !wr_addr[12] && !wr_addr[9];
wire          wr_en_ip = wr_en && !wr_addr[12] &&  wr_addr[9] && !wr_addr[1];
wire          wr_en_sr = wr_en && !wr_addr[12] &&  wr_addr[9] &&  wr_addr[1] && !wr_addr[0];
wire          wr_en_cr = wr_en && !wr_addr[12] &&  wr_addr[9] &&  wr_addr[1] &&  wr_addr[0];

reg           tsc_reset;
reg           tsc_sync;
reg           tsc_power;

wire          tsc_done;
wire    [8:0] tsc_pc;

tsc utsc (.clk(eclk), .reset(tsc_reset), .done(tsc_done), .sync(tsc_sync), .power(tsc_power), .pc(tsc_pc),
    .t_select(t_select), .t_clk_yl(t_clk_yl), .t_reset(t_reset), .t_syl(t_syl), .t_sync_yl(t_sync_yl), .t_sync_yr(t_sync_yr),
	.t_clk_y(t_clk_y), .t_syr(t_syr), .t_sync_y(t_sync_y), .t_clk_yr(t_clk_yr),
	.t_sync_x(t_sync_x), .t_clk_x(t_clk_x), .t_pc(t_pc), .t_shr(t_shr), .t_shs(t_shs),
    .t_mclk(t_mclk), .t_rsmp(t_rsmp), .t_vsmp(t_vsmp),
    .t_ppiclk(t_ppiclk), .t_ppihs(t_ppihs), .t_ppivs(t_ppivs),
    .wclk(clk0), .we(wr_en_uc), .wa(wr_addr[8:0]), .wd(wr_data), .wipe(wr_en_ip), .wip(wr_data[8:0]));

reg           rd_dvld;
reg    [31:0] rd_data;

always @(posedge clk0) begin
    tsc_sync <= wr_en_sr;

    if(wr_en_cr) begin
	tsc_reset <= wr_data[0];
	tsc_power <= wr_data[4];
    end

    rd_dvld <= rd_vld && (rd_addr[14:12] == 3'b010);
    rd_data <= { 7'h00, tsc_pc, 7'h00, tsc_done, 3'h0, tsc_power, 3'h0, tsc_reset };

    if(mrst) begin
	tsc_reset <= 1'b1;
	tsc_sync  <= 1'b0;
	tsc_power <= 1'b0;
    end
end

wire          irq_done = tsc_done;

endmodule
