`define CKDIV_1000 2'b00
`define CKDIV_100  2'b10
`define CKDIV_10   2'b11

module txmac(clk0, mrst, eclk, erst,
    ck_div,
    tx_dvld, tx_data,
    tx_pvld, tx_plen, tx_pflg,
    cr_vld, cr_count,
    e_txclk, e_txen, e_txd);

input         clk0;
input         mrst;
input         eclk;
input         erst;

input   [1:0] ck_div;

input         tx_dvld;
input  [31:0] tx_data;

input         tx_pvld;
input  [13:0] tx_plen;
input   [1:0] tx_pflg;

output        cr_vld;
output [11:0] cr_count;

output        e_txclk;
output        e_txen;
output  [3:0] e_txd;

reg     [5:0] eclkcntr = 6'h00;
wire          eclkzero = !eclkcntr;
reg           eclken, eclksel, eclksel_q, eclkalt = 1'b0, eclkenn;
always @(posedge eclk) begin
    eclkcntr <= eclkcntr + 6'h01;
    if(eclkcntr == (ck_div[0] ? 6'h31 : 6'h04))
        eclkcntr <= 6'h00;

    eclksel_q <= eclksel && !ck_div[0];
    if(!eclkcntr)
        eclksel  <= 1'b1;
    if(eclkcntr == (ck_div[0] ? 6'h19 : 6'h02))
        eclksel  <= 1'b0;

    eclkalt <= eclkzero ^ eclkalt;
    eclken  <= ck_div[1] ? (eclkzero &&  eclkalt) : 1'b1;
    eclkenn <= ck_div[1] ? (eclkzero && !eclkalt) : 1'b0;
end

reg    [11:0] wrptr;
always @(posedge clk0) begin
    if(tx_dvld)
        wrptr <= wrptr + 12'h001;

    if(mrst)
        wrptr <= 12'h000;
end

wire          pf_re, pf_rv;
wire    [1:0] pf_flag;
wire   [13:0] pf_len;
xfifo512x16a pfifo (.wclk(clk0), .wrst(mrst), .we(tx_pvld), .wd({ tx_pflg, tx_plen }),
    .rclk(eclk), .rrst(erst), .re(pf_re), .rv(pf_rv), .rd({ pf_flag, pf_len }));

wire   [13:0] pf_len_4 = tx_plen[13:0] + 14'h0003;
wire   [11:0] pf_words = pf_len_4[13:2];

`define SG_IDLE 2'b00
`define SG_BUSY 2'b01
`define SG_DONE 2'b11
reg     [1:0] sg_state;
assign        pf_re = !sg_state[1] && pf_rv;
reg     [5:0] sg_cntr;
wire    [4:0] sg_wptr = sg_state[0] ? sg_cntr[4:0] : 5'h00;
reg    [11:0] sg_rptr, sg_nptr;
reg    [11:0] sg_cred, sg_cred_eclk;
reg           sg_ctgl_eclk;
reg           sg_ack;
always @(posedge eclk) begin
    if(pf_re) begin
	sg_state <= pf_flag[1] ? `SG_DONE : `SG_BUSY;
	sg_cntr  <= sg_state[0] ? (sg_cntr + 6'h01) : 6'h01;
	sg_nptr  <= sg_nptr + pf_words;
	sg_cred  <= sg_cred + pf_words;
    end

    if(sg_ack) begin
	sg_state <= `SG_IDLE;
	sg_rptr  <= sg_nptr;
	sg_cred_eclk <=  sg_cred;
	sg_ctgl_eclk <= ~sg_ctgl_eclk;
	sg_cred  <= 12'h000;
    end

    if(erst) begin
	sg_state <= `SG_IDLE;
	sg_rptr  <= 12'h000;
	sg_nptr  <= 12'h000;
	sg_cred  <= 12'h000;
	sg_ctgl_eclk <= 1'b0;
    end
end

reg     [4:0] sr_cntr;
wire    [1:0] sr_flag;
wire   [13:0] sr_len;
reg           sr_start, sr_start_q;
reg    [13:0] sr_bcntr;
reg           sr_plast;
reg     [2:0] sr_lastincr;
reg           sr_active;
reg    [13:0] sr_rptr;
wire          sr_next = (sr_bcntr == 14'h0001);
xram32x16 psgram (.clk(eclk), .we(pf_re), .wa(sg_wptr), .wd({ pf_flag, pf_len }), .ra(sr_cntr), .rd({ sr_flag, sr_len }));
always @(posedge eclk) begin
    if(eclken) begin
        if(sr_start)
            sr_active <= 1'b1;
        if(sr_next && sr_plast && sr_active)
            sr_active <= 1'b0;

        sr_start_q <= sr_start;

        if(sr_start || (sr_next && sr_active)) begin
            sr_rptr <= sr_start ? { sg_rptr, 2'b00 } : (sr_rptr + { 11'h000, sr_lastincr });
            sr_plast <= sr_flag[1];
            sr_lastincr <= (sr_len[1:0] == 2'b00) ? 3'b001 : (sr_len[1:0] == 2'b01) ? 3'b100 : (sr_len[1:0] == 2'b10) ? 3'b011 : 3'b010;
            sr_cntr  <= (sr_start || !sr_plast) ? (sr_cntr + 5'h01) : 5'h00;
            sr_bcntr <= sr_len;
        end else if(sr_active) begin
            sr_bcntr <= sr_bcntr - 14'h0001;
            sr_rptr <= sr_rptr + 14'h0001;
        end
    end

    if(erst) begin
        sr_cntr <= 5'h00;
        sr_active <= 1'b0;
    end
end

reg           tpd_vld;
wire    [7:0] tpd_data;
xram4096x32r8qa dram (.wclk(clk0), .we(tx_dvld), .wa(wrptr), .wd(tx_data),
    .rclk(eclk), .re(sr_active && eclken), .ra(sr_rptr), .rd(tpd_data));
always @(posedge eclk)
    if(eclken)
        tpd_vld <= sr_active;

wire   [31:0] tpd_crc32;
crc32_8 txcrc (.clk(eclk), .rst(sr_start_q), .en(tpd_vld && eclken), .i(tpd_data), .q(tpd_crc32));

reg           fpd_start;
reg     [3:0] crc_vld;
reg           tx_busy;
reg     [4:0] fpd_start_q;
always @(posedge eclk) begin
    if(eclken) begin
        fpd_start_q <= { fpd_start_q[3:0], fpd_start };
        sr_start <= fpd_start_q[4];

        if(fpd_start)
            tx_busy <= 1'b1;

        crc_vld  <= { crc_vld[2:0], tpd_vld };
        if(crc_vld[3])
            tx_busy <= 1'b0;
    end

    if(erst)
        tx_busy <= 1'b0;
end

always @(posedge eclk) begin
    if(eclken) begin
        fpd_start <= 1'b0;
        if((sg_state == `SG_DONE) && !fpd_start && !tx_busy)
            fpd_start <= 1'b1;
    end

    sg_ack <= 1'b0;
    if(sr_start && eclken)
        sg_ack <= 1'b1;
end

reg           fpd_vld;
reg     [7:0] fpd_data;
always @(posedge eclk)
    if(eclken) begin
        fpd_vld  <= tpd_vld || crc_vld || sr_active || fpd_start_q || sr_start || fpd_start;
        fpd_data <= (fpd_start_q || sr_start || fpd_start) ? 8'h55 :
                    tpd_vld    ? tpd_data :
                    sr_active  ? 8'hD5 :
                    crc_vld[0] ? ~{ tpd_crc32[24], tpd_crc32[25], tpd_crc32[26], tpd_crc32[27], tpd_crc32[28], tpd_crc32[29], tpd_crc32[30], tpd_crc32[31] } :
                    crc_vld[1] ? ~{ tpd_crc32[16], tpd_crc32[17], tpd_crc32[18], tpd_crc32[19], tpd_crc32[20], tpd_crc32[21], tpd_crc32[22], tpd_crc32[23] } :
                    crc_vld[2] ? ~{ tpd_crc32[8],  tpd_crc32[9],  tpd_crc32[10], tpd_crc32[11], tpd_crc32[12], tpd_crc32[13], tpd_crc32[14], tpd_crc32[15] } :
                    crc_vld[3] ? ~{ tpd_crc32[0],  tpd_crc32[1],  tpd_crc32[2],  tpd_crc32[3],  tpd_crc32[4],  tpd_crc32[5],  tpd_crc32[6],  tpd_crc32[7]  } : 8'h00;
    end else if(eclkenn)
        fpd_data[3:0] <= fpd_data[7:4];

wire    [3:0] hpd_data = ck_div[1] ? fpd_data[3:0] : fpd_data[7:4];

wire          txclk_c0 = ck_div[1] ? ~(eclksel || eclksel_q) : 1'b1;
wire          txclk_c1 = ck_div[1] ? ~eclksel                : 1'b0;

wire          d_txclk;
wire          b_txen;
wire    [3:0] b_txd;
ODDR2 txclk_oddr (.Q(d_txclk),  .C0(eclk), .C1(!eclk), .CE(1'b1), .D0(txclk_c0),    .D1(txclk_c1),    .R(1'b0), .S(1'b0));
ODDR2 txen_oddr  (.Q(b_txen),   .C0(eclk), .C1(!eclk), .CE(1'b1), .D0(fpd_vld),     .D1(fpd_vld),     .R(1'b0), .S(1'b0));
ODDR2 txd0_oddr  (.Q(b_txd[0]), .C0(eclk), .C1(!eclk), .CE(1'b1), .D0(fpd_data[0]), .D1(hpd_data[0]), .R(1'b0), .S(1'b0));
ODDR2 txd1_oddr  (.Q(b_txd[1]), .C0(eclk), .C1(!eclk), .CE(1'b1), .D0(fpd_data[1]), .D1(hpd_data[1]), .R(1'b0), .S(1'b0));
ODDR2 txd2_oddr  (.Q(b_txd[2]), .C0(eclk), .C1(!eclk), .CE(1'b1), .D0(fpd_data[2]), .D1(hpd_data[2]), .R(1'b0), .S(1'b0));
ODDR2 txd3_oddr  (.Q(b_txd[3]), .C0(eclk), .C1(!eclk), .CE(1'b1), .D0(fpd_data[3]), .D1(hpd_data[3]), .R(1'b0), .S(1'b0));
defparam txclk_oddr.DDR_ALIGNMENT = "C0";
defparam txen_oddr.DDR_ALIGNMENT  = "C0";
defparam txd0_oddr.DDR_ALIGNMENT  = "C0";
defparam txd1_oddr.DDR_ALIGNMENT  = "C0";
defparam txd2_oddr.DDR_ALIGNMENT  = "C0";
defparam txd3_oddr.DDR_ALIGNMENT  = "C0";
defparam txclk_oddr.SRTYPE = "ASYNC";
defparam txen_oddr.SRTYPE  = "ASYNC";
defparam txd0_oddr.SRTYPE  = "ASYNC";
defparam txd1_oddr.SRTYPE  = "ASYNC";
defparam txd2_oddr.SRTYPE  = "ASYNC";
defparam txd3_oddr.SRTYPE  = "ASYNC";

wire          b_txclk;
IODELAY2 txclk_iodelay2 (
    .DATAOUT(), .DATAOUT2(), .DOUT(b_txclk), .TOUT(),
    .BUSY(), .CAL(), .CE(), .INC(), .CLK(), .RST(),
    .IDATAIN(), .ODATAIN(d_txclk), .T(),
    .IOCLK0(), .IOCLK1());
defparam txclk_iodelay2.DELAY_SRC = "ODATAIN";
defparam txclk_iodelay2.ODELAY_VALUE = 20;
defparam txclk_iodelay2.IDELAY_TYPE = "FIXED";

OBUF  txclk_obuf (.O(e_txclk),  .I(b_txclk));
OBUF  txen_obuf  (.O(e_txen),   .I(b_txen));
OBUF  txd0_obuf  (.O(e_txd[0]), .I(b_txd[0]));
OBUF  txd1_obuf  (.O(e_txd[1]), .I(b_txd[1]));
OBUF  txd2_obuf  (.O(e_txd[2]), .I(b_txd[2]));
OBUF  txd3_obuf  (.O(e_txd[3]), .I(b_txd[3]));

reg           sg_ctgl_clk0;
reg           sg_ctgl_clk0_q;

reg           cr_vld;
reg    [11:0] cr_count;
always @(posedge clk0) begin
    sg_ctgl_clk0   <= sg_ctgl_eclk;
    sg_ctgl_clk0_q <= sg_ctgl_clk0;

    cr_vld   <= sg_ctgl_clk0 ^ sg_ctgl_clk0_q;
    cr_count <= sg_cred_eclk;

    if(mrst)
        sg_ctgl_clk0 <= 1'b0;
end

endmodule
