module xfifo256x24(clk, rst, we, wd, re, rv, rd);

input         clk;
input         rst;

input         we;
input  [23:0] wd;

input         re;
output        rv;
output [23:0] rd;

reg     [8:0] wa, wa_q;
always @(posedge clk) begin
    wa_q <= wa;
    if(we)
	wa <= wa + 9'h001;
    if(rst)
	wa <= 9'h000;
end

wire    [7:0] dummy;
wire   [23:0] rd_int;

wire          re_int;
reg     [8:0] ra;

RAMB8BWER ram (.DOADO({dummy, rd_int[23:16]}), .DOBDO(rd_int[15:0]), .DOPADOP(), .DOPBDOP(), 
    .ADDRAWRADDR({wa[7:0], 5'h00}), .ADDRBRDADDR({ra[7:0], 5'h00}),
    .CLKAWRCLK(clk), .CLKBRDCLK(clk),
    .DIADI({8'h00, wd[23:16]}), .DIBDI(wd[15:0]), .DIPADIP(2'h0), .DIPBDIP(2'h0),
    .ENAWREN(we), .ENBRDEN(re_int), .REGCEA(1'b1), .REGCEBREGCE(1'b1),
    .RSTA(rst), .RSTBRST(rst), .WEAWEL({2{we}}), .WEBWEU({2{we}}));
defparam ram.RAM_MODE = "SDP";
defparam ram.DATA_WIDTH_A = 36;
defparam ram.DATA_WIDTH_B = 36;
defparam ram.DOA_REG = 1;
defparam ram.DOB_REG = 1;

reg           rv;
reg    [23:0] rd;

reg           re_q, re_qq;
assign        re_int = !rv && !re_q && !re_qq && (ra != wa_q);

always @(posedge clk) begin
    re_q <= re_int;
    re_qq <= re_q;

    if(re_int)
	ra <= ra + 9'h001;

    if(re_qq) begin
	rv <= 1'b1;
	rd <= rd_int;
    end

    if(re && rv)
	rv <= 1'b0;

    if(rst) begin
	ra <= 9'h000;
	rv <= 1'b0;
    end
end

endmodule
