module xfifo256x40(clk, rst, we, wd, re, rv, rd);

input         clk;
input         rst;

input         we;
input  [39:0] wd;

input         re;
output        rv;
output [39:0] rd;

reg           we_q, we_qq;
reg     [8:0] wa, wa_q;
reg    [39:0] wd_q;
reg    [19:0] wd_qq;
always @(posedge clk) begin
    we_q  <= we;
    we_qq <= we_q;
    if(we)
	wd_q  <= wd;
    if(we_q)
	wd_qq <= wd_q[39:20];

    wa_q <= wa;
    if(we_q)
	wa <= wa + 9'h001;
    if(rst)
	wa <= 9'h000;
end

wire          we_fold = we_q || we_qq;
wire    [8:0] wa_fold = { wa_q[7:0], we_qq };
wire   [19:0] wd_fold = we_qq ? wd_qq : wd_q[19:0];

wire          re_int;
reg     [8:0] ra;
reg           rs, rs_q, rs_qq;
wire          re_fold = re_int || rs;
wire    [8:0] ra_fold = { ra[8:0], rs };
wire   [19:0] rd_fold;

wire   [11:0] dummy;
RAMB16BWER ram (
    .DOA(), .DOB({dummy, rd_fold}), .DOPA(), .DOPB(),
    .ADDRA({wa_fold, 5'h00}), .ADDRB({ra_fold, 5'h00}),
    .CLKA(clk), .CLKB(clk),
    .DIA({12'h000, wd_fold}), .DIB(32'h00000000), .DIPA(4'h0), .DIPB(4'h0),
    .ENA(we_fold), .ENB(re_fold), .REGCEA(1'b0), .REGCEB(1'b1),
    .RSTA(rst), .RSTB(rst), .WEA({4{we_fold}}), .WEB(4'h0));
defparam ram.DATA_WIDTH_A = 36;
defparam ram.DATA_WIDTH_B = 36;
defparam ram.DOB_REG = 1;
defparam ram.SIM_DEVICE = "SPARTAN6";

reg           rv;
reg    [39:0] rd;

assign        re_int = (ra != wa_q) && !rs && !rs_q && !rs_qq;

always @(posedge clk) begin
    if(rs)
	ra <= ra + 9'h001;
    rs    <= re_int;
    rs_q  <= rs;
    rs_qq <= rs_q;

    if(rs_q)
	rd[19:0]  <= rd_fold;
    if(rs_qq) begin
	rd[39:20] <= rd_fold;
	rv <= 1'b1;
    end

    if(rv && re)
	rv <= 1'b0;

    if(rst) begin
	ra <= 9'h000;
	rv <= 1'b0;
    end
end

endmodule
