module xfifo512x16a(wclk, wrst, we, wd, rclk, rrst, re, rv, rd);

input         wclk;
input         wrst;
input         we;
input  [15:0] wd;

input         rclk;
input         rrst;
input         re;
output        rv;
output [15:0] rd;

wire          re_int;
wire   [15:0] rd_int;

reg     [9:0] wa;
reg     [9:0] ra;

wire   [15:0] dummy;
RAMB16BWER ram (
    .DOA(), .DOB({dummy, rd_int}), .DOPA(), .DOPB(),
    .ADDRA({1'b0, wa[8:0], 4'h0}), .ADDRB({1'b0, ra[8:0], 4'h0}),
    .CLKA(wclk), .CLKB(rclk),
    .DIA({16'h000, wd}), .DIB(32'h00000000), .DIPA(4'h0), .DIPB(4'h0),
    .ENA(we), .ENB(re_int), .REGCEA(1'b0), .REGCEB(1'b1),
    .RSTA(wrst), .RSTB(rrst), .WEA({4{we}}), .WEB(4'h0));
defparam ram.DATA_WIDTH_A = 18;
defparam ram.DATA_WIDTH_B = 18;
defparam ram.DOB_REG = 1;
defparam ram.SIM_DEVICE = "SPARTAN6";

wire    [9:0] wa_n = wa + 10'h001;
wire    [9:0] wg_n = wa_n ^ {1'b0, wa_n[9:1]};

reg     [9:0] wg;

always @(posedge wclk) begin
    if(we) begin
	wa <= wa_n;
	wg <= wg_n;
    end

    if(wrst) begin
	wa <= 10'h000;
	wg <= 10'h000;
    end
end

wire    [9:0] ra_n = ra + 10'h001;
wire    [9:0] rg_n = ra_n ^ {1'b0, ra_n[9:1]};

reg     [9:0] wg_s1, wg_s2;
reg     [9:0] rg;

wire          rv_int = (rg != wg_s2);

always @(posedge rclk) begin
    wg_s1 <= wg;
    wg_s2 <= wg_s1;

    if(re_int) begin
	ra <= ra_n;
	rg <= rg_n;
    end

    if(rrst) begin
	ra <= 10'h000;
	rg <= 10'h000;
    end
end

reg           rv;
reg    [15:0] rd;

reg           re_int_q, re_int_qq;
assign        re_int = !rv && !re_int_q && !re_int_qq && rv_int;

always @(posedge rclk) begin
    re_int_q <= re_int;
    re_int_qq <= re_int_q;

    if(re_int_qq) begin
	rv <= 1'b1;
	rd <= rd_int;
    end

    if(re && rv)
	rv <= 1'b0;

    if(rrst)
	rv <= 1'b0;
end

endmodule
