#include <stdio.h>
#include <string.h>

/* GR RE
   BL GR */
unsigned short raw[4560][3052];
unsigned short dark[4560][3052];
unsigned char img[4560][3052][3];

static unsigned short bswap16(unsigned short x)
{
    return (x >> 8) | (x << 8);
}

void load_deswizzle(char *fn)
{
    FILE *f = fopen(fn, "rb");
    unsigned short row[3052];
    unsigned x, y;

    fscanf(f, "%*s%*d%*d%*d%*c");

    for(y=0; y<4560; y++) {
	fread(row, 3052, 2, f);
	for(x=0; x<3052; x++)
	    raw[y][x^(y&1)] = bswap16(row[(x & ~3)|((x & 2) >> 1)|((x & 1) << 1)]);
    }

    fclose(f);
}

void save_raw(char *fn)
{
    FILE *f = fopen(fn, "wb");
    unsigned short row[3052];
    unsigned x, y;

    fprintf(f, "P5\n3052 4560\n65535\n");
    for(y=0; y<4560; y++) {
	for(x=0; x<3052; x++)
	    row[x] = bswap16(raw[y][x]);
	fwrite(row, 3052, 2, f);
    }
    fclose(f);
}

void save_img(char *fn)
{
    FILE *f = fopen(fn, "wb");
    fprintf(f, "P6\n3052 4560\n255\n");
    fwrite(img, 4560, 3052*3, f);
    fclose(f);
}

void normalize(void)
{
    unsigned x, y;
    unsigned short max;
    unsigned long long scale;

    max = 0;
    for(y=0; y<4560; y++)
	for(x=0; x<3052; x++) {
	    if(raw[y][x] > dark[y][x])
		raw[y][x] -= dark[y][x];
	    else
		raw[y][x] = 0;
	    if(y > 40 && x > 40 && y < 4520 && x < 3012)
		if(raw[y][x] > max)
		    max = raw[y][x];
	}

    scale = 0xFFFFFFFF / max;

    for(y=0; y<4560; y++)
	for(x=0; x<3052; x++) {
	    if(raw[y][x] > max)
		raw[y][x] = 65535;
	    else
		raw[y][x] = (raw[y][x] * scale) >> 16;
	}
}

void debayer(void)
{
    unsigned x, y;

    for(y=2; y<4558; y++)
	for(x=2; x<3050; x++) {
	    /* G in R row */
	    if(!(x & 1) && !(y & 1)) {
		img[y][x][0] = (10*raw[y][x] + 8*(raw[y][x-1] + raw[y][x+1]) - 2*(raw[y-1][x-1] + raw[y-1][x+1] + raw[y][x-2] + raw[y][x+2] + raw[y+1][x-1] + raw[y+1][x+1]) + (raw[y-2][x] + raw[y+2][x])) >> 12;
		img[y][x][1] = raw[y][x] >> 8;
		img[y][x][2] = (10*raw[y][x] + 8*(raw[y-1][x] + raw[y+1][x]) - 2*(raw[y-1][x-1] + raw[y-1][x+1] + raw[y-2][x] + raw[y+2][x] + raw[y+1][x-1] + raw[y+1][x+1]) + (raw[y][x-2] + raw[y][x+2])) >> 12;
	    }

	    /* R */
	    if((x & 1) && !(y & 1)) {
		img[y][x][0] = raw[y][x] >> 8;
		img[y][x][1] = (4*raw[y][x] + 2*(raw[y-1][x] + raw[y+1][x] + raw[y][x-1] + raw[y][x+1]) - (raw[y-2][x] + raw[y+2][x] + raw[y][x-2] + raw[y][x+2])) >> 11;
		img[y][x][2] = (12*raw[y][x] + 4*(raw[y-1][x-1] + raw[y-1][x+1] + raw[y+1][x-1] + raw[y+1][x+1]) - 3*(raw[y-2][x] + raw[y+2][x] + raw[y][x-2] + raw[y][x+2])) >> 12;
	    }

	    /* G in B row */
	    if((x & 1) && (y & 1)) {
		img[y][x][0] = (10*raw[y][x] + 8*(raw[y-1][x] + raw[y+1][x]) - 2*(raw[y-1][x-1] + raw[y-1][x+1] + raw[y-2][x] + raw[y+2][x] + raw[y+1][x-1] + raw[y+1][x+1]) + (raw[y][x-2] + raw[y][x+2])) >> 12;
		img[y][x][1] = raw[y][x] >> 8;
		img[y][x][2] = (10*raw[y][x] + 8*(raw[y][x-1] + raw[y][x+1]) - 2*(raw[y-1][x-1] + raw[y-1][x+1] + raw[y][x-2] + raw[y][x+2] + raw[y+1][x-1] + raw[y+1][x+1]) + (raw[y-2][x] + raw[y+2][x])) >> 12;
	    }

	    /* B */
	    if(!(x & 1) && (y & 1)) {
		img[y][x][0] = (12*raw[y][x] + 4*(raw[y-1][x-1] + raw[y-1][x+1] + raw[y+1][x-1] + raw[y+1][x+1]) - 3*(raw[y-2][x] + raw[y+2][x] + raw[y][x-2] + raw[y][x+2])) >> 12;
		img[y][x][1] = (4*raw[y][x] + 2*(raw[y-1][x] + raw[y+1][x] + raw[y][x-1] + raw[y][x+1]) - (raw[y-2][x] + raw[y+2][x] + raw[y][x-2] + raw[y][x+2])) >> 11;
		img[y][x][2] = raw[y][x] >> 8;
	    }
	}
}

int main(void)
{
    load_deswizzle("darkframe.pgm");
    memcpy(dark, raw, sizeof(raw));
    load_deswizzle("capture.pgm");

    normalize();
    debayer();

    save_img("capture.ppm");

    return 0;
}
