#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <micser.h>

#define MAXLINE	1024
#define STACK   16
#define MAXDIR  256
#define MAXFILE	4096
#define MAXMIME	64
#define MAXLINK 256

char links[MAXLINK][MS_FNAME_SIZE];
unsigned nlinks = 0;
unsigned link_pack(char *str)
{
    unsigned i;
    for(i=0; i<nlinks; i++)
	if(!strcmp(str, links[i]))
	    return i;
    strcpy(links[i], str);
    nlinks++;
    return i;
}

char mimes[MAXMIME][MS_FNAME_SIZE];
unsigned mimeoffs[MAXMIME], mimesize = 0;
unsigned nmimes = 0;
unsigned mime_pack(char *str)
{
    unsigned i;
    for(i=0; i<nmimes; i++)
	if(!strcmp(str, mimes[i]))
	    return mimeoffs[i];
    strcpy(mimes[i], str);
    mimeoffs[i] = mimesize;
    mimesize += strlen(str) + 1;
    nmimes++;
    return mimeoffs[i];
}

void set_defent(ms_file *dir, char *str)
{
    unsigned i;

    if(!str[0])
	return;

    for(i=0; i<dir->dir.count; i++)
	if(!strcmp(str, (char *)dir->dir.entries[i].name))
	    break;
    if(i >= dir->dir.count) {
	fprintf(stderr, "Error: default entry '%s' not found.\n", str);
	exit(1);
    }

    dir->dir.defent = i;
}

void hexwrite(void *_buf, unsigned length, FILE *f, unsigned *offset)
{
    unsigned char *buf = _buf;
    unsigned i, p;

    p = *offset;
    for(i=0; i<length; i++) {
	if(p && !(p & 31))
	    fprintf(f, ",\n   ");
	if(p & 31)
	    fprintf(f, ", 0x%02X", buf[i]);
	else
	    fprintf(f, "0x%02X", buf[i]);
	p++;
    }
    *offset = p;
}

ms_file *files[MAXFILE];
unsigned foffset[MAXFILE], flength[MAXFILE];
unsigned nfiles = 0;

int main(void)
{
    FILE *layout, *srcfile, *packfile;
    char line[MAXLINE], *p, *name, *source, *mime;
    unsigned indent, stack[STACK], depth=0;
    unsigned type, partial, length, i, j, offset;
    char defent[STACK][MS_FNAME_SIZE];
    ms_file *root=NULL, *file, *dirstack[STACK];

    stack[0] = 0;

    chdir("test");
    layout = fopen("LAYOUT", "r");
    while(fgets(line, sizeof(line), layout)) {
	/* trim end of line whitespace */
	p = line + strlen(line);
	while(p != line && strchr(" \t\r\n", p[-1]))
	    *(--p) = 0;
	if(!line[0])
	    continue;

	/* count start of line whitespace */
	indent = 0;
	for(p=line; *p && strchr(" \t", *p); p++)
	    if(*p == '\t')
		indent = (indent + 8) & ~7;
	    else
		indent++;
	name = p;

	/* figure out stack depth */
	if(indent > stack[depth]) {
	    depth ++;
	    stack[depth] = indent;
	} else while(indent < stack[depth] && depth) {
	    depth --;
	    set_defent(dirstack[depth], defent[depth]);
	}

	/* figure out type */
	p = strpbrk(name, "\t ");
	if(p)
	    *(p++) = 0;

	type = MS_FT_STATIC;
	if(name[strlen(name)-1] == '/') {
	    type = MS_FT_DIRECTORY;
	    name[strlen(name)-1] = 0;
	}

	mime = source = NULL;
	partial = 0;

	if(type != MS_FT_DIRECTORY) {
	    if(!p) {
		fprintf(stderr, "Error: file '%s' has missing tags.\n", name);
		exit(1);
	    }

	    while(*p && strchr("\t ", *p))
		p++;
	    source = p;

	    p = strpbrk(source, "\t ");
	    if(!p) {
		fprintf(stderr, "Error: file '%s' has missing tags.\n", name);
		exit(1);
	    }
	    *(p++) = 0;

	    while(*p && strchr("\t ", *p))
		p++;
	    mime = p;

	    if(source[strlen(source)-1] == ')') {
		type = MS_FT_GENERATED;
		if(source[strlen(source)-2] == '+') {
		    partial = 1;
		    if(source[strlen(source)-3] != '(') {
			fprintf(stderr, "Error: file '%s' has mismatched parens.\n", name);
			exit(1);
		    }
		    source[strlen(source)-3] = 0;
		} else if(source[strlen(source)-2] != '(') {
		    fprintf(stderr, "Error: file '%s' has mismatched parens.\n", name);
		    exit(1);
		} else
		    source[strlen(source)-2] = 0;
	    }
	} else {
	    if(p) {
		while(*p && strchr("\t ", *p))
		    p++;
		strcpy(defent[depth], p);
	    } else
		defent[depth][0] = 0;
	}

	if(!root && (type != MS_FT_DIRECTORY || name[0] || depth)) {
	    fprintf(stderr, "Error: layout must start with a root directory ('/').\n");
	    exit(1);
	}

	if(root && !depth) {
	    fprintf(stderr, "Error: only the root directory can be at level 0.\n");
	    exit(1);
	}

	file = NULL;

	switch(type) {
	case MS_FT_DIRECTORY:
	    file = malloc(12 + MAXDIR*(MS_FNAME_SIZE+4));
	    file->type = type;
	    file->dir.count = 0;
	    file->dir.defent = 0xFFFFFFFF;
	    if(!root)
		root = file;
	    dirstack[depth] = file;
	    break;
	case MS_FT_STATIC:
	    srcfile = fopen(source, "rb");
	    if(!srcfile) {
		fprintf(stderr, "Error: file '%s' source '%s' is missing.\n", name, source);
		exit(1);
	    }
	    fseek(srcfile, 0, SEEK_END);
	    length = ftell(srcfile);
	    fseek(srcfile, 0, SEEK_SET);
	    file = malloc(12 + ((length+3)&~3));
	    file->type = type;
	    file->st.mime = mime_pack(mime);
	    file->st.size = length;
	    fread(file->st.data, length, 1, srcfile);
	    fclose(srcfile);
	    break;
	case MS_FT_GENERATED:
	    file = malloc(16);
	    file->type = type;
	    file->gen.mime = mime_pack(mime);
	    file->gen.partial = partial;
	    file->gen.link = link_pack(source);
	    break;
	}

	if(file != root) {
	    strcpy((char *)dirstack[depth-1]->dir.entries[dirstack[depth-1]->dir.count].name, name);
	    dirstack[depth-1]->dir.entries[dirstack[depth-1]->dir.count].ptr = nfiles;
	    dirstack[depth-1]->dir.count++;
	}
	files[nfiles++] = file;
    }
    fclose(layout);

    while(depth) {
	depth --;
	set_defent(dirstack[depth], defent[depth]);
    }

    offset = 0;
    for(i=0; i<nfiles; i++) {
	foffset[i] = offset;
	length = 0;
	switch(files[i]->type) {
	case MS_FT_DIRECTORY:
	    length = 12 + files[i]->dir.count*(MS_FNAME_SIZE+4);
	    break;
	case MS_FT_STATIC:
	    length = 12 + files[i]->st.size;
	    break;
	case MS_FT_GENERATED:
	    length = 16;
	}
	length = (length + 3) & ~3;
	flength[i] = length;
	offset += length;
    }

    for(i=0; i<nfiles; i++)
	switch(files[i]->type) {
	case MS_FT_DIRECTORY:
	    for(j=0; j<files[i]->dir.count; j++)
		files[i]->dir.entries[j].ptr = foffset[files[i]->dir.entries[j].ptr];
	    break;
	case MS_FT_STATIC:
	    files[i]->st.mime += offset;
	    break;
	case MS_FT_GENERATED:
	    files[i]->gen.mime += offset;
	}

    packfile = fopen("fs.h", "w");
    fprintf(packfile, "/* auto-generated file */\n\n");

    fprintf(packfile, "static u8 fs_data[%d] = {\n   ", offset+mimesize);

    offset = 0;
    for(i=0; i<nfiles; i++)
	hexwrite(files[i], flength[i], packfile, &offset);
    for(i=0; i<nmimes; i++)
	hexwrite(mimes[i], strlen(mimes[i])+1, packfile, &offset);

    fprintf(packfile, " };\n\n");

    if(nlinks) {
	fprintf(packfile, "static ms_gen_func fs_links[%d] = {\n", nlinks);
	for(i=0; i<nlinks; i++) {
	    fprintf(packfile, "    %s", links[i]);
	    if(i == nlinks-1)
		fprintf(packfile, " };\n");
	    else
		fprintf(packfile, ",\n");
	}
    } else
	fprintf(packfile, "#define fs_links ((ms_gen_func *)0)\n");

    fclose(packfile);

    return 0;
}
