#define _BSD_SOURCE 1
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

float clk = 1.0f;

char *bits[] = {
    "vsmp",
    "rsmp",
    "clk_x",
    "sync_x",
    "syr",
    "syl",
    "shr",
    "reset",
    "select",
    "shs",
    "pc",
    "sync_yr",
    "sync_y",
    "sync_yl",
    "clk_yr",
    "clk_y",
    "clk_yl",
    "ppihs",
    "ppivs",
    "done",
};

int main(void)
{
    FILE *f = fopen("14mp.asm", "r");
    char line[1024], *l, *p, *q;
    char *latx[512];
    int ladr[512];
    int nl = 0;
    char *rftx[512];
    int radr[512];
    int nr = 0;
    float dly;
    int dlycnt, i, j, addr=0;
    int cstate = 0;
    int ops[512];
    memset(ops, 0, sizeof(ops));
    while(fgets(line, 1024, f)) {
	p = strpbrk(line, "\n\r;");
	if(p)
	    *p = 0;
	for(l=line; *l==' ' || *l=='\t'; l++) ;
	if(!*l)
	    continue;
	for(p=l; *p; p++)
	    if(*p == '\t')
		*p = ' ';
	p = strchr(l, ':');
	if(p) {
	    *p=0;
	    q = strchr(l, ' ');
	    if(q)
		*q=0;
	    fprintf(stderr, "Label <%s> at %d.\n", l, addr);
	    ladr[nl] = addr;
	    latx[nl++] = strdup(l);
	    l=p+1;
	    while(*l==' ')
		l++;
	    if(!*l)
		continue;
	}
	if(!strncmp(l, "clk ", 4)) {
	    l = l+4;
	    while(*l==' ')
		l++;
	    clk = atof(l);
	    fprintf(stderr, "Set clock to %f.\n", clk);
	    continue;
	}
	if(!strncmp(l, "sync ", 5)) {
	    l = l+5;
	    while(*l==' ')
		l++;
	    i = atoi(l);
	    fprintf(stderr, "Sync to source %d.\n", i);
	    ops[addr++] = 0x30000000|i;
	    continue;
	}
	if(!strncmp(l, "loop0 ", 6) || !strncmp(l, "loop1 ", 6)) {
	    j = l[4]-'0';
	    l = l+5;
	    while(*l==' ')
		l++;
	    p=strchr(l, ' ');
	    if(!p) {
		fprintf(stderr, "Wut?\n");
		return 1;
	    }
	    *p=0;
	    p++;
	    while(*p==' ')
		p++;
	    i = atoi(p);
	    radr[nr] = addr;
	    rftx[nr++] = strdup(l);
	    fprintf(stderr, "Loop to <%s> at %d (count %d).\n", l, addr, i);
	    ops[addr++] = 0x10000000|(i-1)|(j<<15);
	    continue;
	}
	if(!strncmp(l, "jump ", 5)) {
	    l = l+5;
	    while(*l==' ')
		l++;
	    p=strchr(l, ' ');
	    if(p)
		*p = 0;
	    radr[nr] = addr;
	    rftx[nr++] = strdup(l);
	    fprintf(stderr, "Jump to <%s> at %d.\n", l, addr);
	    ops[addr++] = 0x20000000;
	    continue;
	}
	if(!strcmp(l, "nop")) {
	    ops[addr++] = cstate;
	    continue;
	}
	p = strchr(l, '+');
	if(p) {
	    *p = 0;
	    p++;
	    while(*p==' ')
		p++;
	    dly = atof(p);
	    dlycnt = (int)roundf(dly / clk);
	    fprintf(stderr, "Rounded %f delay to %d cycle%s.\n", dly, dlycnt, dlycnt==1?"":"s");
	    if(dlycnt < 1)
		dlycnt = 0;
	    else
		dlycnt--;
	} else {
	    dlycnt = 0;
	}
	while(*l) {
	    p = strchr(l, '=');
	    if(!p) {
		fprintf(stderr, "Huh (%s)?\n", l);
		return 1;
	    }
	    *p=0;
	    p++;
	    for(i=0; i<20; i++)
		if(!strcmp(l, bits[i]))
		    break;
	    if(i==20) {
		fprintf(stderr, "Huh <%s>?\n", l);
		return 1;
	    }
	    if(*p == '1') {
		cstate |= 1<<i;
	    } else if(*p == '0') {
		cstate &= ~(1<<i);
	    } else {
		fprintf(stderr, "Huh [%c]?\n", *p);
		return 1;
	    }
	    l = p+1;
	    while(*l==' ')
		l++;
	}
	while(dlycnt > 255) {
	    ops[addr++] = cstate|(255<<20);
	    dlycnt -= 255;
	}
	ops[addr++] = cstate|(dlycnt<<20);
    }
    fclose(f);

    for(i=0; i<nr; i++) {
	for(j=0; j<nl; j++)
	    if(!strcmp(rftx[i], latx[j]))
		break;
	if(j==nl) {
	    fprintf(stderr, "Huh {%s}?\n", rftx[i]);
	    return 1;
	}
	ops[radr[i]] |= ladr[j]<<16;
	fprintf(stderr, "Reference <%s> at %d resolved to %d.\n", rftx[i], radr[i], ladr[j]);
    }

    f=fopen("tscrom.txt" ,"w");
    fprintf(f, "@000\n");
    for(i=0; i<512; i+=8)
	for(j=0; j<8; j++) {
	    fprintf(f, "%08X", ops[i+j]);
	    if(j == 7)
		fprintf(f, "\n");
	    else
		fprintf(f, " ");
	}
    fclose(f);

    addr = (addr + 7) & ~7;

    f=fopen("tscrom.h" ,"w");
    fprintf(f, "#define TSC_ROM_SIZE %d\n", addr);
    fprintf(f, "static unsigned int tsc_rom[TSC_ROM_SIZE] = {\n");
    for(i=0; i<addr; i+=8) {
	fprintf(f, "    ");
	for(j=0; j<8; j++) {
	    fprintf(f, "0x%08X", ops[i+j]);
	    if((i+j) < addr-1)
		fprintf(f, ",");
	    if(j<7)
		fprintf(f, " ");
	}
	if((i+j) == addr)
	    fprintf(f, " };");
	fprintf(f, "\n");
    }
    fclose(f);

    return 0;
}
